# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Integration tests for :func:`iris_grib._save_rules.grid_definition_section`
to confirm that the correct grid_definition_template is being selected.

"""

# Import iris_grib.tests first so that some things can be initialised before
# importing anything else.
import iris_grib.tests as tests

from iris.coords import DimCoord
from iris.coord_systems import (GeogCS,
                                RotatedGeogCS,
                                Mercator,
                                TransverseMercator,
                                LambertConformal,
                                AlbersEqualArea)
from iris.cube import Cube
import numpy as np

from iris_grib._save_rules import grid_definition_section
from iris_grib.tests.unit.save_rules import GdtTestMixin


class Test(tests.IrisGribTest, GdtTestMixin):
    def setUp(self):
        GdtTestMixin.setUp(self)
        self.ellipsoid = GeogCS(6371200)

    def test_grid_definition_template_0(self):
        # Regular lat/lon (Plate Carree).
        x_points = np.arange(3)
        y_points = np.arange(3)
        coord_units = 'degrees'
        cs = self.ellipsoid
        test_cube = self._make_test_cube(cs, x_points, y_points, coord_units)
        grid_definition_section(test_cube, self.mock_grib)
        self._check_key('gridDefinitionTemplateNumber', 0)

    def test_grid_definition_template_1(self):
        # Rotated lat/lon (Plate Carree).
        x_points = np.arange(3)
        y_points = np.arange(3)
        coord_units = 'degrees'
        cs = RotatedGeogCS(34.0, 117.0, ellipsoid=self.ellipsoid)
        test_cube = self._make_test_cube(cs, x_points, y_points, coord_units)
        grid_definition_section(test_cube, self.mock_grib)
        self._check_key('gridDefinitionTemplateNumber', 1)

    def test_grid_definition_template_4(self):
        # Irregular (variable resolution) lat/lon grid.
        x_points = np.array([0, 2, 7])
        y_points = np.array([1, 3, 6])
        coord_units = '1'
        cs = self.ellipsoid
        test_cube = self._make_test_cube(cs, x_points, y_points, coord_units)
        grid_definition_section(test_cube, self.mock_grib)
        self._check_key('gridDefinitionTemplateNumber', 4)

    def test_grid_definition_template_5(self):
        # Irregular (variable resolution) rotated lat/lon grid.
        x_points = np.array([0, 2, 7])
        y_points = np.array([1, 3, 6])
        coord_units = '1'
        cs = RotatedGeogCS(34.0, 117.0, ellipsoid=self.ellipsoid)
        test_cube = self._make_test_cube(cs, x_points, y_points, coord_units)
        grid_definition_section(test_cube, self.mock_grib)
        self._check_key('gridDefinitionTemplateNumber', 5)

    def test_grid_definition_template_10(self):
        # Mercator grid.
        x_points = np.arange(3)
        y_points = np.arange(3)
        coord_units = 'm'
        cs = Mercator(ellipsoid=self.ellipsoid)
        test_cube = self._make_test_cube(cs, x_points, y_points, coord_units)
        grid_definition_section(test_cube, self.mock_grib)
        self._check_key('gridDefinitionTemplateNumber', 10)

    def test_grid_definition_template_12(self):
        # Transverse Mercator grid.
        x_points = np.arange(3)
        y_points = np.arange(3)
        coord_units = 'm'
        cs = TransverseMercator(0, 0, 0, 0, 1, ellipsoid=self.ellipsoid)
        test_cube = self._make_test_cube(cs, x_points, y_points, coord_units)
        grid_definition_section(test_cube, self.mock_grib)
        self._check_key('gridDefinitionTemplateNumber', 12)

    def test_grid_definition_template_30(self):
        # Lambert Conformal grid.
        x_points = np.arange(3)
        y_points = np.arange(3)
        coord_units = 'm'
        cs = LambertConformal(ellipsoid=self.ellipsoid)
        test_cube = self._make_test_cube(cs, x_points, y_points, coord_units)
        grid_definition_section(test_cube, self.mock_grib)
        self._check_key('gridDefinitionTemplateNumber', 30)

    def test_coord_system_not_supported(self):
        # Test an unsupported grid - let's choose Albers Equal Area.
        x_points = np.arange(3)
        y_points = np.arange(3)
        coord_units = '1'
        cs = AlbersEqualArea(ellipsoid=self.ellipsoid)
        test_cube = self._make_test_cube(cs, x_points, y_points, coord_units)

        exp_name = cs.grid_mapping_name.replace('_', ' ').title()
        exp_emsg = 'not supported for coordinate system {!r}'.format(exp_name)
        with self.assertRaisesRegex(ValueError, exp_emsg):
            grid_definition_section(test_cube, self.mock_grib)


if __name__ == "__main__":
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Test function :func:`iris_grib._load_convert.scanning_mode.

"""

# import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests

from iris.exceptions import TranslationError

from iris_grib._load_convert import scanning_mode, ScanningMode


class Test(tests.IrisGribTest):
    def test_unset(self):
        expected = ScanningMode(False, False, False, False)
        self.assertEqual(scanning_mode(0x0), expected)

    def test_i_negative(self):
        expected = ScanningMode(i_negative=True, j_positive=False,
                                j_consecutive=False, i_alternative=False)
        self.assertEqual(scanning_mode(0x80), expected)

    def test_j_positive(self):
        expected = ScanningMode(i_negative=False, j_positive=True,
                                j_consecutive=False, i_alternative=False)
        self.assertEqual(scanning_mode(0x40), expected)

    def test_j_consecutive(self):
        expected = ScanningMode(i_negative=False, j_positive=False,
                                j_consecutive=True, i_alternative=False)
        self.assertEqual(scanning_mode(0x20), expected)

    def test_i_alternative(self):
        with self.assertRaises(TranslationError):
            scanning_mode(0x10)


if __name__ == '__main__':
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Tests for function :func:`iris_grib._load_convert._hindcast_fix`.

"""

# import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests

from collections import namedtuple

from iris_grib._load_convert import _hindcast_fix as hindcast_fix


class TestHindcastFix(tests.IrisGribTest):
    # setup tests : provided value, fix-applies, expected-fixed
    FixTest = namedtuple('FixTest', ('given', 'fixable', 'fixed'))
    test_values = [
        FixTest(0, False, None),
        FixTest(100, False, None),
        FixTest(2 * 2**30 - 1, False, None),
        FixTest(2 * 2**30, False, None),
        FixTest(2 * 2**30 + 1, True, -1),
        FixTest(2 * 2**30 + 2, True, -2),
        FixTest(3 * 2**30 - 1, True, -(2**30 - 1)),
        FixTest(3 * 2**30, False, None)]

    def setUp(self):
        self.patch_warn = self.patch('warnings.warn')

    def test_fix(self):
        # Check hindcast fixing.
        for given, fixable, fixed in self.test_values:
            result = hindcast_fix(given)
            expected = fixed if fixable else given
            self.assertEqual(result, expected)

    def test_fix_warning(self):
        # Check warning appears when enabled.
        self.patch('iris_grib._load_convert.options.warn_on_unsupported', True)
        hindcast_fix(2 * 2**30 + 5)
        self.assertEqual(self.patch_warn.call_count, 1)
        self.assertIn('Re-interpreting large grib forecastTime',
                      self.patch_warn.call_args[0][0])

    def test_fix_warning_disabled(self):
        # Default is no warning.
        hindcast_fix(2 * 2**30 + 5)
        self.assertEqual(self.patch_warn.call_count, 0)


if __name__ == '__main__':
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Grib save implementation.

:mod:`iris_grib._save_rules` is a private module with no public API.
It is invoked from :meth:`iris_grib.save_grib2`.

"""

import warnings

import cf_units
import gribapi
import numpy as np
import numpy.ma as ma
import cartopy.crs as ccrs

import iris
from iris.aux_factory import HybridHeightFactory, HybridPressureFactory
from iris.coord_systems import (GeogCS, RotatedGeogCS, Mercator,
                                TransverseMercator, LambertConformal)
from iris.exceptions import TranslationError


from ._iris_mercator_support import confirm_extended_mercator_supported
from . import grib_phenom_translation as gptx
from ._load_convert import (_STATISTIC_TYPE_NAMES, _TIME_RANGE_UNITS,
                            _SPATIAL_PROCESSING_TYPES)
from .grib_phenom_translation import GRIBCode
from iris.util import is_regular, regular_step


# Invert code tables from :mod:`iris_grib._load_convert`.
_STATISTIC_TYPE_NAMES = {val: key for key, val in
                         _STATISTIC_TYPE_NAMES.items()}
_TIME_RANGE_UNITS = {val: key for key, val in _TIME_RANGE_UNITS.items()}


def fixup_float32_as_int32(value):
    """
    Workaround for use when the ECMWF GRIB API treats an IEEE 32-bit
    floating-point value as a signed, 4-byte integer.

    Returns the integer value which will result in the on-disk
    representation corresponding to the IEEE 32-bit floating-point
    value.

    """
    value_as_float32 = np.array(value, dtype='f4')
    value_as_uint32 = value_as_float32.view(dtype='u4')
    if value_as_uint32 >= 0x80000000:
        # Convert from two's-complement to sign-and-magnitude.
        # NB. Because of the silly representation of negative
        # integers in GRIB2, there is no value we can pass to
        # grib_set that will result in the bit pattern 0x80000000.
        # But since that bit pattern corresponds to a floating
        # point value of negative-zero, we can safely treat it as
        # positive-zero instead.
        value_as_grib_int = 0x80000000 - int(value_as_uint32)
    else:
        value_as_grib_int = int(value_as_uint32)
    return value_as_grib_int


def fixup_int32_as_uint32(value):
    """
    Workaround for use when the ECMWF GRIB API treats a signed, 4-byte
    integer value as an unsigned, 4-byte integer.

    Returns the unsigned integer value which will result in the on-disk
    representation corresponding to the signed, 4-byte integer value.

    """
    value = int(value)
    if -0x7fffffff <= value <= 0x7fffffff:
        if value < 0:
            # Convert from two's-complement to sign-and-magnitude.
            value = 0x80000000 - value
    else:
        msg = '{} out of range -2147483647 to 2147483647.'.format(value)
        raise ValueError(msg)
    return value


def ensure_set_int32_value(grib, key, value):
    """
    Ensure the workaround function :func:`fixup_int32_as_uint32` is applied as
    necessary to problem keys.

    """
    try:
        gribapi.grib_set(grib, key, value)
    except gribapi.GribInternalError:
        value = fixup_int32_as_uint32(value)
        gribapi.grib_set(grib, key, value)


###############################################################################
#
# Constants
#
###############################################################################

# Reference Flag Table 3.3
_RESOLUTION_AND_COMPONENTS_GRID_WINDS_BIT = 3  # NB "bit5", from MSB=1.

# Reference Regulation 92.1.6
_DEFAULT_DEGREES_UNITS = 1.0e-6


###############################################################################
#
# Identification Section 1
#
###############################################################################


def centre(cube, grib):
    # TODO: read centre from cube
    gribapi.grib_set_long(grib, "centre", 74)  # UKMO
    gribapi.grib_set_long(grib, "subCentre", 0)  # exeter is not in the spec


def reference_time(cube, grib):
    # Set the reference time.
    # (analysis, forecast start, verify time, obs time, etc)
    try:
        fp_coord = cube.coord("forecast_period")
    except iris.exceptions.CoordinateNotFoundError:
        fp_coord = None

    if fp_coord is not None:
        rt, rt_meaning, _, _ = _non_missing_forecast_period(cube)
    else:
        rt, rt_meaning, _, _ = _missing_forecast_period(cube)

    gribapi.grib_set_long(grib, "significanceOfReferenceTime", rt_meaning)
    gribapi.grib_set_long(
        grib, "dataDate", "%04d%02d%02d" % (rt.year, rt.month, rt.day))
    gribapi.grib_set_long(
        grib, "dataTime", "%02d%02d" % (rt.hour, rt.minute))

    # TODO: Set the calendar, when we find out what happened to the proposal!
    # http://tinyurl.com/oefqgv6
    # I was sure it was approved for pre-operational use but it's not there.


def identification(cube, grib):
    centre(cube, grib)
    reference_time(cube, grib)

    # operational product, operational test, research product, etc
    # (missing for now)
    gribapi.grib_set_long(grib, "productionStatusOfProcessedData", 255)

    # Code table 1.4
    # analysis, forecast, processed satellite, processed radar,
    if cube.coords('realization'):
        # assume realization will always have 1 and only 1 point
        # as cubes saving to GRIB2 a 2D horizontal slices
        if cube.coord('realization').points[0] != 0:
            gribapi.grib_set_long(grib, "typeOfProcessedData", 4)
        else:
            gribapi.grib_set_long(grib, "typeOfProcessedData", 3)
    else:
        gribapi.grib_set_long(grib, "typeOfProcessedData", 2)


###############################################################################
#
# Grid Definition Section 3
#
###############################################################################


def shape_of_the_earth(cube, grib):
    # assume latlon
    cs = cube.coord(dimensions=[0]).coord_system

    # Initially set shape_of_earth keys to missing (255 for byte, -1 for long).
    gribapi.grib_set_long(grib, "scaleFactorOfRadiusOfSphericalEarth", 255)
    gribapi.grib_set_long(grib, "scaledValueOfRadiusOfSphericalEarth", -1)
    gribapi.grib_set_long(grib, "scaleFactorOfEarthMajorAxis", 255)
    gribapi.grib_set_long(grib, "scaledValueOfEarthMajorAxis", -1)
    gribapi.grib_set_long(grib, "scaleFactorOfEarthMinorAxis", 255)
    gribapi.grib_set_long(grib, "scaledValueOfEarthMinorAxis", -1)

    if isinstance(cs, GeogCS):
        ellipsoid = cs
    else:
        ellipsoid = cs.ellipsoid
        if ellipsoid is None:
            msg = "Could not determine shape of the earth from coord system "\
                  "of horizontal grid."
            raise TranslationError(msg)

    # Spherical earth.
    if ellipsoid.inverse_flattening == 0.0:
        gribapi.grib_set_long(grib, "shapeOfTheEarth", 1)
        gribapi.grib_set_long(grib, "scaleFactorOfRadiusOfSphericalEarth", 0)
        gribapi.grib_set_long(grib, "scaledValueOfRadiusOfSphericalEarth",
                              ellipsoid.semi_major_axis)
        gribapi.grib_set_long(grib, "scaleFactorOfEarthMajorAxis", 0)
        gribapi.grib_set_long(grib, "scaledValueOfEarthMajorAxis", 0)
        gribapi.grib_set_long(grib, "scaleFactorOfEarthMinorAxis", 0)
        gribapi.grib_set_long(grib, "scaledValueOfEarthMinorAxis", 0)
    # Oblate spheroid earth.
    else:
        gribapi.grib_set_long(grib, "shapeOfTheEarth", 7)
        gribapi.grib_set_long(grib, "scaleFactorOfEarthMajorAxis", 0)
        gribapi.grib_set_long(grib, "scaledValueOfEarthMajorAxis",
                              ellipsoid.semi_major_axis)
        gribapi.grib_set_long(grib, "scaleFactorOfEarthMinorAxis", 0)
        gribapi.grib_set_long(grib, "scaledValueOfEarthMinorAxis",
                              ellipsoid.semi_minor_axis)


def grid_dims(x_coord, y_coord, grib, x_str, y_str):
    gribapi.grib_set_long(grib, x_str, x_coord.shape[0])
    gribapi.grib_set_long(grib, y_str, y_coord.shape[0])


def latlon_first_last(x_coord, y_coord, grib):
    if x_coord.has_bounds() or y_coord.has_bounds():
        warnings.warn("Ignoring xy bounds")

# XXX Pending #1125
#    gribapi.grib_set_double(grib, "latitudeOfFirstGridPointInDegrees",
#                            float(y_coord.points[0]))
#    gribapi.grib_set_double(grib, "latitudeOfLastGridPointInDegrees",
#                            float(y_coord.points[-1]))
#    gribapi.grib_set_double(grib, "longitudeOfFirstGridPointInDegrees",
#                            float(x_coord.points[0]))
#    gribapi.grib_set_double(grib, "longitudeOfLastGridPointInDegrees",
#                            float(x_coord.points[-1]))
# WORKAROUND
    gribapi.grib_set_long(grib, "latitudeOfFirstGridPoint",
                          int(y_coord.points[0]*1000000))
    gribapi.grib_set_long(grib, "latitudeOfLastGridPoint",
                          int(y_coord.points[-1]*1000000))
    gribapi.grib_set_long(grib, "longitudeOfFirstGridPoint",
                          int((x_coord.points[0] % 360)*1000000))
    gribapi.grib_set_long(grib, "longitudeOfLastGridPoint",
                          int((x_coord.points[-1] % 360)*1000000))


def dx_dy(x_coord, y_coord, grib):
    x_step = regular_step(x_coord)
    y_step = regular_step(y_coord)
    # Set x and y step.  For degrees, this is encoded as an integer:
    # 1 * 10^6 * floating point value.
    # WMO Manual on Codes regulation 92.1.6
    if x_coord.units == 'degrees':
        gribapi.grib_set(grib, "iDirectionIncrement",
                         round(1e6 * float(abs(x_step))))
    else:
        raise ValueError('X coordinate must be in degrees, not {}'
                         '.'.format(x_coord.units))
    if y_coord.units == 'degrees':
        gribapi.grib_set(grib, "jDirectionIncrement",
                         round(1e6 * float(abs(y_step))))
    else:
        raise ValueError('Y coordinate must be in degrees, not {}'
                         '.'.format(y_coord.units))


def scanning_mode_flags(x_coord, y_coord, grib):
    gribapi.grib_set_long(grib, "iScansPositively",
                          int(x_coord.points[1] - x_coord.points[0] > 0))
    gribapi.grib_set_long(grib, "jScansPositively",
                          int(y_coord.points[1] - y_coord.points[0] > 0))


def horizontal_grid_common(cube, grib, xy=False):
    nx_str, ny_str = ("Nx", "Ny") if xy else ("Ni", "Nj")
    # Grib encoding of the sequences of X and Y points.
    y_coord = cube.coord(dimensions=[0])
    x_coord = cube.coord(dimensions=[1])
    shape_of_the_earth(cube, grib)
    grid_dims(x_coord, y_coord, grib, nx_str, ny_str)
    scanning_mode_flags(x_coord, y_coord, grib)


def latlon_points_regular(cube, grib):
    y_coord = cube.coord(dimensions=[0])
    x_coord = cube.coord(dimensions=[1])
    latlon_first_last(x_coord, y_coord, grib)
    dx_dy(x_coord, y_coord, grib)


def latlon_points_irregular(cube, grib):
    y_coord = cube.coord(dimensions=[0])
    x_coord = cube.coord(dimensions=[1])

    # Distinguish between true-north and grid-oriented vectors.
    is_grid_wind = cube.name() in ('x_wind', 'y_wind', 'grid_eastward_wind',
                                   'grid_northward_wind')
    # Encode in bit "5" of 'resolutionAndComponentFlags' (other bits unused).
    component_flags = 0
    if is_grid_wind:
        component_flags |= 2 ** _RESOLUTION_AND_COMPONENTS_GRID_WINDS_BIT
    gribapi.grib_set(grib, 'resolutionAndComponentFlags', component_flags)

    # Record the  X and Y coordinate values.
    # NOTE: there is currently a bug in the gribapi which means that the size
    # of the longitudes array does not equal 'Nj', as it should.
    # See : https://software.ecmwf.int/issues/browse/SUP-1096
    # So, this only works at present if the x and y dimensions are **equal**.
    lon_values = x_coord.points / _DEFAULT_DEGREES_UNITS
    lat_values = y_coord.points / _DEFAULT_DEGREES_UNITS
    gribapi.grib_set_array(grib, 'longitudes',
                           np.array(np.round(lon_values), dtype=np.int64))
    gribapi.grib_set_array(grib, 'latitudes',
                           np.array(np.round(lat_values), dtype=np.int64))


def rotated_pole(cube, grib):
    # Grib encoding of a rotated pole coordinate system.
    cs = cube.coord(dimensions=[0]).coord_system

    if cs.north_pole_grid_longitude != 0.0:
        raise TranslationError(
            'Grib save does not yet support Rotated-pole coordinates with '
            'a rotated prime meridian.')
# XXX Pending #1125
#    gribapi.grib_set_double(grib, "latitudeOfSouthernPoleInDegrees",
#                            float(cs.n_pole.latitude))
#    gribapi.grib_set_double(grib, "longitudeOfSouthernPoleInDegrees",
#                            float(cs.n_pole.longitude))
#    gribapi.grib_set_double(grib, "angleOfRotationInDegrees", 0)
# WORKAROUND
    latitude = cs.grid_north_pole_latitude / _DEFAULT_DEGREES_UNITS
    longitude = (((cs.grid_north_pole_longitude + 180) % 360) /
                 _DEFAULT_DEGREES_UNITS)
    gribapi.grib_set(grib, "latitudeOfSouthernPole", - int(round(latitude)))
    gribapi.grib_set(grib, "longitudeOfSouthernPole", int(round(longitude)))
    gribapi.grib_set(grib, "angleOfRotation", 0)


def points_in_unit(coord, unit):
    points = coord.units.convert(coord.points, unit)
    points = np.around(points).astype(int)
    return points


def step(points, atol):
    diffs = points[1:] - points[:-1]
    mean_diff = np.mean(diffs).astype(points.dtype)
    if not np.allclose(diffs, mean_diff, atol=atol):
        raise ValueError()
    return int(mean_diff)


def grid_definition_template_0(cube, grib):
    """
    Set keys within the provided grib message based on
    Grid Definition Template 3.0.

    Template 3.0 is used to represent "latitude/longitude (or equidistant
    cylindrical, or Plate Carree)".
    The coordinates are regularly spaced, true latitudes and longitudes.

    """
    # Constant resolution, aka 'regular' true lat-lon grid.
    gribapi.grib_set_long(grib, "gridDefinitionTemplateNumber", 0)
    horizontal_grid_common(cube, grib)
    latlon_points_regular(cube, grib)


def grid_definition_template_1(cube, grib):
    """
    Set keys within the provided grib message based on
    Grid Definition Template 3.1.

    Template 3.1 is used to represent "rotated latitude/longitude (or
    equidistant cylindrical, or Plate Carree)".
    The coordinates are regularly spaced, rotated latitudes and longitudes.

    """
    # Constant resolution, aka 'regular' rotated lat-lon grid.
    gribapi.grib_set_long(grib, "gridDefinitionTemplateNumber", 1)

    # Record details of the rotated coordinate system.
    rotated_pole(cube, grib)

    # Encode the lat/lon points.
    horizontal_grid_common(cube, grib)
    latlon_points_regular(cube, grib)


def grid_definition_template_4(cube, grib):
    """
    Set keys within the provided grib message based on
    Grid Definition Template 3.4.

    Template 3.4 is used to represent "variable resolution latitude/longitude".
    The coordinates are irregularly spaced latitudes and longitudes.

    """
    # XXX: will we need to set `Ni` and `Nj`?
    gribapi.grib_set(grib, "gridDefinitionTemplateNumber", 4)
    horizontal_grid_common(cube, grib)
    latlon_points_irregular(cube, grib)


def grid_definition_template_5(cube, grib):
    """
    Set keys within the provided grib message based on
    Grid Definition Template 3.5.

    Template 3.5 is used to represent "variable resolution rotated
    latitude/longitude".
    The coordinates are irregularly spaced, rotated latitudes and longitudes.

    """
    # NOTE: we must set Ni=Nj=1 before establishing the template.
    # Without this, setting "gridDefinitionTemplateNumber" = 5 causes an
    # immediate error.
    # See: https://software.ecmwf.int/issues/browse/SUP-1095
    # This is acceptable, as the subsequent call to 'horizontal_grid_common'
    # will set these to the correct horizontal dimensions
    # (by calling 'grid_dims').
    gribapi.grib_set(grib, "Ni", 1)
    gribapi.grib_set(grib, "Nj", 1)
    gribapi.grib_set(grib, "gridDefinitionTemplateNumber", 5)

    # Record details of the rotated coordinate system.
    rotated_pole(cube, grib)
    # Encode the lat/lon points.
    horizontal_grid_common(cube, grib)
    latlon_points_irregular(cube, grib)


def grid_definition_template_10(cube, grib):
    """
    Set keys within the provided grib message based on
    Grid Definition Template 3.10.

    Template 3.10 is used to represent a Mercator grid.

    """
    gribapi.grib_set(grib, "gridDefinitionTemplateNumber", 10)

    # Retrieve some information from the cube.
    y_coord = cube.coord(dimensions=[0])
    x_coord = cube.coord(dimensions=[1])
    cs = y_coord.coord_system

    # Normalise the coordinate values to millimetres - the resolution
    # used in the GRIB message.
    y_mm = points_in_unit(y_coord, 'mm')
    x_mm = points_in_unit(x_coord, 'mm')

    # Encode the horizontal points.

    # NB. Since we're already in millimetres, our tolerance for
    # discrepancy in the differences is 1.
    try:
        x_step = step(x_mm, atol=1)
        y_step = step(y_mm, atol=1)
    except ValueError:
        msg = 'Irregular coordinates not supported for Mercator.'
        raise TranslationError(msg)

    gribapi.grib_set(grib, 'Di', abs(x_step))
    gribapi.grib_set(grib, 'Dj', abs(y_step))

    horizontal_grid_common(cube, grib)

    # Transform first and last points into geographic CS.
    geog = cs.ellipsoid if cs.ellipsoid is not None else GeogCS(1)
    first_x, first_y, = geog.as_cartopy_crs().transform_point(
        x_coord.points[0],
        y_coord.points[0],
        cs.as_cartopy_crs())
    last_x, last_y = geog.as_cartopy_crs().transform_point(
        x_coord.points[-1],
        y_coord.points[-1],
        cs.as_cartopy_crs())
    first_x = first_x % 360
    last_x = last_x % 360

    gribapi.grib_set(grib, "latitudeOfFirstGridPoint",
                     int(np.round(first_y / _DEFAULT_DEGREES_UNITS)))
    gribapi.grib_set(grib, "longitudeOfFirstGridPoint",
                     int(np.round(first_x / _DEFAULT_DEGREES_UNITS)))
    gribapi.grib_set(grib, "latitudeOfLastGridPoint",
                     int(np.round(last_y / _DEFAULT_DEGREES_UNITS)))
    gribapi.grib_set(grib, "longitudeOfLastGridPoint",
                     int(np.round(last_x / _DEFAULT_DEGREES_UNITS)))

    # Check and raise a more intelligible error, if the Iris version is too old
    # to support the Mercator 'standard_parallel' property.
    confirm_extended_mercator_supported()
    # Encode the latitude at which the projection intersects the Earth.
    gribapi.grib_set(grib, 'LaD',
                     cs.standard_parallel / _DEFAULT_DEGREES_UNITS)

    # Encode resolution and component flags
    gribapi.grib_set(grib, 'resolutionAndComponentFlags',
                     0x1 << _RESOLUTION_AND_COMPONENTS_GRID_WINDS_BIT)


def grid_definition_template_12(cube, grib):
    """
    Set keys within the provided grib message based on
    Grid Definition Template 3.12.

    Template 3.12 is used to represent a Transverse Mercator grid.

    """
    gribapi.grib_set(grib, "gridDefinitionTemplateNumber", 12)

    # Retrieve some information from the cube.
    y_coord = cube.coord(dimensions=[0])
    x_coord = cube.coord(dimensions=[1])
    cs = y_coord.coord_system

    # Normalise the coordinate values to centimetres - the resolution
    # used in the GRIB message.
    y_cm = points_in_unit(y_coord, 'cm')
    x_cm = points_in_unit(x_coord, 'cm')

    # Set some keys specific to GDT12.
    # Encode the horizontal points.

    # NB. Since we're already in centimetres, our tolerance for
    # discrepancy in the differences is 1.
    try:
        x_step = step(x_cm, atol=1)
        y_step = step(y_cm, atol=1)
    except ValueError:
        msg = ('Irregular coordinates not supported for Transverse '
               'Mercator.')
        raise TranslationError(msg)
    gribapi.grib_set(grib, 'Di', abs(x_step))
    gribapi.grib_set(grib, 'Dj', abs(y_step))
    horizontal_grid_common(cube, grib)

    # GRIBAPI expects unsigned ints in X1, X2, Y1, Y2 but it should accept
    # signed ints, so work around this.
    # See https://software.ecmwf.int/issues/browse/SUP-1101
    ensure_set_int32_value(grib, 'Y1', int(y_cm[0]))
    ensure_set_int32_value(grib, 'Y2', int(y_cm[-1]))
    ensure_set_int32_value(grib, 'X1', int(x_cm[0]))
    ensure_set_int32_value(grib, 'X2', int(x_cm[-1]))

    # Lat and lon of reference point are measured in millionths of a degree.
    gribapi.grib_set(grib, "latitudeOfReferencePoint",
                     cs.latitude_of_projection_origin / _DEFAULT_DEGREES_UNITS)
    gribapi.grib_set(grib, "longitudeOfReferencePoint",
                     cs.longitude_of_central_meridian / _DEFAULT_DEGREES_UNITS)

    # Convert a value in metres into the closest integer number of
    # centimetres.
    def m_to_cm(value):
        return int(round(value * 100))

    # False easting and false northing are measured in units of (10^-2)m.
    gribapi.grib_set(grib, 'XR', m_to_cm(cs.false_easting))
    gribapi.grib_set(grib, 'YR', m_to_cm(cs.false_northing))

    # GRIBAPI expects a signed int for scaleFactorAtReferencePoint
    # but it should accept a float, so work around this.
    # See https://software.ecmwf.int/issues/browse/SUP-1100
    value = cs.scale_factor_at_central_meridian
    key_type = gribapi.grib_get_native_type(grib,
                                            "scaleFactorAtReferencePoint")
    if key_type is not float:
        value = fixup_float32_as_int32(value)
    gribapi.grib_set(grib, "scaleFactorAtReferencePoint", value)


def grid_definition_template_30(cube, grib):
    """
    Set keys within the provided grib message based on
    Grid Definition Template 3.30.

    Template 3.30 is used to represent a Lambert Conformal grid.

    """

    gribapi.grib_set(grib, "gridDefinitionTemplateNumber", 30)

    # Retrieve some information from the cube.
    y_coord = cube.coord(dimensions=[0])
    x_coord = cube.coord(dimensions=[1])
    cs = y_coord.coord_system

    # Normalise the coordinate values to millimetres - the resolution
    # used in the GRIB message.
    y_mm = points_in_unit(y_coord, 'mm')
    x_mm = points_in_unit(x_coord, 'mm')

    # Encode the horizontal points.

    # NB. Since we're already in millimetres, our tolerance for
    # discrepancy in the differences is 1.
    try:
        x_step = step(x_mm, atol=1)
        y_step = step(y_mm, atol=1)
    except ValueError:
        msg = ('Irregular coordinates not supported for Lambert '
               'Conformal.')
        raise TranslationError(msg)
    gribapi.grib_set(grib, 'Dx', abs(x_step))
    gribapi.grib_set(grib, 'Dy', abs(y_step))

    horizontal_grid_common(cube, grib, xy=True)

    # Transform first point into geographic CS
    geog = cs.ellipsoid if cs.ellipsoid is not None else GeogCS(1)
    first_x, first_y = geog.as_cartopy_crs().transform_point(
        x_coord.points[0],
        y_coord.points[0],
        cs.as_cartopy_crs())
    first_x = first_x % 360
    central_lon = cs.central_lon % 360

    gribapi.grib_set(grib, "latitudeOfFirstGridPoint",
                     int(np.round(first_y / _DEFAULT_DEGREES_UNITS)))
    gribapi.grib_set(grib, "longitudeOfFirstGridPoint",
                     int(np.round(first_x / _DEFAULT_DEGREES_UNITS)))
    gribapi.grib_set(grib, "LaD", cs.central_lat / _DEFAULT_DEGREES_UNITS)
    gribapi.grib_set(grib, "LoV", central_lon / _DEFAULT_DEGREES_UNITS)
    latin1, latin2 = cs.secant_latitudes
    gribapi.grib_set(grib, "Latin1", latin1 / _DEFAULT_DEGREES_UNITS)
    gribapi.grib_set(grib, "Latin2", latin2 / _DEFAULT_DEGREES_UNITS)
    gribapi.grib_set(grib, 'resolutionAndComponentFlags',
                     0x1 << _RESOLUTION_AND_COMPONENTS_GRID_WINDS_BIT)

    # Which pole are the parallels closest to? That is the direction
    # that the cone converges.
    poliest_sec = latin1 if abs(latin1) > abs(latin2) else latin2
    centre_flag = 0x0 if poliest_sec > 0 else 0x1
    gribapi.grib_set(grib, 'projectionCentreFlag', centre_flag)
    gribapi.grib_set(grib, "latitudeOfSouthernPole", 0)
    gribapi.grib_set(grib, "longitudeOfSouthernPole", 0)


def grid_definition_section(cube, grib):
    """
    Set keys within the grid definition section of the provided grib message,
    based on the properties of the cube.

    """
    x_coord = cube.coord(dimensions=[1])
    y_coord = cube.coord(dimensions=[0])
    cs = x_coord.coord_system  # N.B. already checked same cs for x and y.
    regular_x_and_y = is_regular(x_coord) and is_regular(y_coord)

    if isinstance(cs, GeogCS):
        if regular_x_and_y:
            grid_definition_template_0(cube, grib)
        else:
            grid_definition_template_4(cube, grib)

    elif isinstance(cs, RotatedGeogCS):
        # Rotated coordinate system cases.
        # Choose between GDT 3.1 and 3.5 according to coordinate regularity.
        if regular_x_and_y:
            grid_definition_template_1(cube, grib)
        else:
            grid_definition_template_5(cube, grib)

    elif isinstance(cs, Mercator):
        # Mercator coordinate system (template 3.10)
        grid_definition_template_10(cube, grib)

    elif isinstance(cs, TransverseMercator):
        # Transverse Mercator coordinate system (template 3.12).
        grid_definition_template_12(cube, grib)

    elif isinstance(cs, LambertConformal):
        # Lambert Conformal coordinate system (template 3.30).
        grid_definition_template_30(cube, grib)
    else:
        name = cs.grid_mapping_name.replace('_', ' ').title()
        emsg = 'Grib saving is not supported for coordinate system {!r}'
        raise ValueError(emsg.format(name))


###############################################################################
#
# Product Definition Section 4
#
###############################################################################

def set_discipline_and_parameter(cube, grib):
    # Default values for parameter identity keys = effectively "MISSING".
    discipline, category, number = 255, 255, 255
    identity_found = False

    # First, see if we can find and interpret a 'GRIB_PARAM' attribute.
    attr = cube.attributes.get('GRIB_PARAM', None)
    if attr:
        try:
            # Convert to standard tuple-derived form.
            gc = GRIBCode(attr)
            if gc.edition == 2:
                discipline = gc.discipline
                category = gc.category
                number = gc.number
                identity_found = True
        except:
            pass

    if not identity_found:
        # Else, translate a cube phenomenon, if possible.
        # NOTE: for now, can match by *either* standard_name or long_name.
        # This allows workarounds for data with no identified standard_name.
        grib2_info = gptx.cf_phenom_to_grib2_info(cube.standard_name,
                                                  cube.long_name)
        if grib2_info is not None:
            discipline = grib2_info.discipline
            category = grib2_info.category
            number = grib2_info.number
            identity_found = True

    if not identity_found:
        warnings.warn('Unable to determine Grib2 parameter code for cube.\n'
                      'discipline, parameterCategory and parameterNumber '
                      'have been set to "missing".')

    gribapi.grib_set(grib, "discipline", discipline)
    gribapi.grib_set(grib, "parameterCategory", category)
    gribapi.grib_set(grib, "parameterNumber", number)


def _non_missing_forecast_period(cube):
    # Calculate "model start time" to use as the reference time.
    fp_coord = cube.coord("forecast_period")

    # Convert fp and t to hours so we can subtract to calculate R.
    cf_fp_hrs = fp_coord.units.convert(fp_coord.points[0], 'hours')
    t_coord = cube.coord("time").copy()
    hours_since = cf_units.Unit("hours since epoch",
                                calendar=t_coord.units.calendar)
    t_coord.convert_units(hours_since)

    rt_num = t_coord.points[0] - cf_fp_hrs
    rt = hours_since.num2date(rt_num)
    rt_meaning = 1  # "start of forecast"

    # Forecast period
    if fp_coord.units == cf_units.Unit("hours"):
        grib_time_code = 1
    elif fp_coord.units == cf_units.Unit("minutes"):
        grib_time_code = 0
    elif fp_coord.units == cf_units.Unit("seconds"):
        grib_time_code = 13
    else:
        raise TranslationError(
            "Unexpected units for 'forecast_period' : %s" % fp_coord.units)

    if not t_coord.has_bounds():
        fp = fp_coord.points[0]
    else:
        if not fp_coord.has_bounds():
            raise TranslationError(
                "bounds on 'time' coordinate requires bounds on"
                " 'forecast_period'.")
        fp = fp_coord.bounds[0][0]

    if fp - int(fp):
        warnings.warn("forecast_period encoding problem: "
                      "scaling required.")
    fp = int(fp)

    return rt, rt_meaning, fp, grib_time_code


def _missing_forecast_period(cube):
    """
    Returns a reference time and significance code together with a forecast
    period and corresponding units type code.

    """
    t_coord = cube.coord("time")

    if cube.coords('forecast_reference_time'):
        # Make copies and convert them to common "hours since" units.
        hours_since = cf_units.Unit('hours since epoch',
                                    calendar=t_coord.units.calendar)
        frt_coord = cube.coord('forecast_reference_time').copy()
        frt_coord.convert_units(hours_since)
        t_coord = t_coord.copy()
        t_coord.convert_units(hours_since)
        # Extract values.
        t = t_coord.bounds[0, 0] if t_coord.has_bounds() else t_coord.points[0]
        frt = frt_coord.points[0]
        # Calculate GRIB parameters.
        rt = frt_coord.units.num2date(frt)
        rt_meaning = 1  # Forecast reference time.
        fp = t - frt
        integer_fp = int(fp)
        if integer_fp != fp:
            msg = 'Truncating floating point forecast period {} to ' \
                  'integer value {}'
            warnings.warn(msg.format(fp, integer_fp))
        fp = integer_fp
        fp_meaning = 1  # Hours
    else:
        # With no forecast period or forecast reference time set assume a
        # reference time significance of "Observation time" and set the
        # forecast period to 0h.
        t = t_coord.bounds[0, 0] if t_coord.has_bounds() else t_coord.points[0]
        rt = t_coord.units.num2date(t)
        rt_meaning = 3  # Observation time
        fp = 0
        fp_meaning = 1  # Hours

    return rt, rt_meaning, fp, fp_meaning


def set_forecast_time(cube, grib):
    """
    Set the forecast time keys based on the forecast_period coordinate. In
    the absence of a forecast_period and forecast_reference_time,
    the forecast time is set to zero.

    """
    try:
        fp_coord = cube.coord("forecast_period")
    except iris.exceptions.CoordinateNotFoundError:
        fp_coord = None

    if fp_coord is not None:
        _, _, fp, grib_time_code = _non_missing_forecast_period(cube)
    else:
        _, _, fp, grib_time_code = _missing_forecast_period(cube)

    gribapi.grib_set(grib, "indicatorOfUnitOfTimeRange", grib_time_code)
    gribapi.grib_set(grib, "forecastTime", fp)


def set_fixed_surfaces(cube, grib, full3d_cube=None):

    # Look for something we can export
    v_coord = grib_v_code = output_unit = None

    # Detect factories for hybrid vertical coordinates.
    hybrid_factories = [
        factory for factory in cube.aux_factories
        if isinstance(factory, (HybridHeightFactory, HybridPressureFactory))]
    if not hybrid_factories:
        hybrid_factory = None
    elif len(hybrid_factories) > 1:
        msg = 'Data contains >1 vertical coordinate factory : {}'
        raise ValueError(msg.format(hybrid_factories))
    else:
        factory = hybrid_factories[0]
        # Fetch the matching 'complete' factory from the *full* 3d cube, so we
        # have all the level information.
        hybrid_factory = full3d_cube.aux_factory(factory.name())

    # Handle various different styles of vertical coordinate.
    # hybrid height / pressure
    if hybrid_factory is not None:
        # N.B. in this case, there are additional operations, besides just
        # encoding v_coord : see below at end ..
        v_coord = cube.coord('model_level_number')
        output_unit = cf_units.Unit("1")
        if isinstance(hybrid_factory, HybridHeightFactory):
            grib_v_code = 118
        elif isinstance(hybrid_factory, HybridPressureFactory):
            grib_v_code = 119
        else:
            msg = 'Unrecognised factory type : {}'
            raise ValueError(msg.format(hybrid_factory))

    # pressure
    elif cube.coords("air_pressure") or cube.coords("pressure"):
        grib_v_code = 100
        output_unit = cf_units.Unit("Pa")
        v_coord = (cube.coords("air_pressure") or cube.coords("pressure"))[0]

    # altitude
    elif cube.coords("altitude"):
        grib_v_code = 102
        output_unit = cf_units.Unit("m")
        v_coord = cube.coord("altitude")

    # height
    elif cube.coords("height"):
        grib_v_code = 103
        output_unit = cf_units.Unit("m")
        v_coord = cube.coord("height")

    # depth
    elif cube.coords("depth"):
        grib_v_code = 106
        output_unit = cf_units.Unit('m')
        v_coord = cube.coord("depth")

    elif cube.coords("air_potential_temperature"):
        grib_v_code = 107
        output_unit = cf_units.Unit('K')
        v_coord = cube.coord("air_potential_temperature")

    # unknown / absent
    else:
        # check for *ANY* height coords at all...
        v_coords = cube.coords(axis='z')
        if v_coords:
            # There are vertical coordinate(s), but we don't understand them...
            v_coords_str = ' ,'.join(["'{}'".format(c.name())
                                      for c in v_coords])
            raise TranslationError(
                'The vertical-axis coordinate(s) ({}) '
                'are not recognised or handled.'.format(v_coords_str))

    # What did we find?
    if v_coord is None:
        # No vertical coordinate: record as 'surface' level (levelType=1).
        # NOTE: may *not* be truly correct, but seems to be common practice.
        # Still under investigation :
        # See https://github.com/SciTools/iris/issues/519
        gribapi.grib_set(grib, "typeOfFirstFixedSurface", 1)
        gribapi.grib_set(grib, "scaleFactorOfFirstFixedSurface", 0)
        gribapi.grib_set(grib, "scaledValueOfFirstFixedSurface", 0)
        # Set secondary surface = 'missing'.
        gribapi.grib_set(grib, "typeOfSecondFixedSurface", -1)
        gribapi.grib_set(grib, "scaleFactorOfSecondFixedSurface", 255)
        gribapi.grib_set(grib, "scaledValueOfSecondFixedSurface", -1)
    elif not v_coord.has_bounds():
        # No second surface
        output_v = v_coord.units.convert(v_coord.points[0], output_unit)
        if output_v - abs(output_v):
            warnings.warn("Vertical level encoding problem: scaling required.")
        output_v = int(round(output_v))

        gribapi.grib_set(grib, "typeOfFirstFixedSurface", grib_v_code)
        gribapi.grib_set(grib, "scaleFactorOfFirstFixedSurface", 0)
        gribapi.grib_set(grib, "scaledValueOfFirstFixedSurface", output_v)
        gribapi.grib_set(grib, "typeOfSecondFixedSurface", -1)
        gribapi.grib_set(grib, "scaleFactorOfSecondFixedSurface", 255)
        gribapi.grib_set(grib, "scaledValueOfSecondFixedSurface", -1)
    else:
        # bounded : set lower+upper surfaces
        output_v = v_coord.units.convert(v_coord.bounds[0], output_unit)
        if output_v[0] - abs(output_v[0]) or output_v[1] - abs(output_v[1]):
            warnings.warn("Vertical level encoding problem: scaling required.")
        gribapi.grib_set(grib, "typeOfFirstFixedSurface", grib_v_code)
        gribapi.grib_set(grib, "typeOfSecondFixedSurface", grib_v_code)
        gribapi.grib_set(grib, "scaleFactorOfFirstFixedSurface", 0)
        gribapi.grib_set(grib, "scaleFactorOfSecondFixedSurface", 0)
        gribapi.grib_set(grib, "scaledValueOfFirstFixedSurface",
                         int(round(output_v[0])))
        gribapi.grib_set(grib, "scaledValueOfSecondFixedSurface",
                         int(round(output_v[1])))

    if hybrid_factory is not None:
        # Need to record ALL the level coefficents in a 'PV' vector.
        level_delta_coord = hybrid_factory.delta
        sigma_coord = hybrid_factory.sigma
        model_levels = full3d_cube.coord('model_level_number').points
        # Just check these make some kind of sense (!)
        if model_levels.dtype.kind not in 'iu':
            msg = 'model_level_number is not an integer: dtype={}.'
            raise ValueError(msg.format(model_levels.dtype))
        if np.min(model_levels) < 1:
            msg = 'model_level_number must be > 0: mininum value = {}.'
            raise ValueError(msg.format(np.min(model_levels)))
        # Need to save enough levels for indexes up to  [max(model_levels)]
        n_levels = np.max(model_levels)
        max_valid_nlevels = 9999
        if n_levels > max_valid_nlevels:
            msg = ('model_level_number values are > {} : '
                   'maximum value = {}.')
            raise ValueError(msg.format(max_valid_nlevels, n_levels))
        # In sample data we have seen, there seems to be an extra missing data
        # value *before* each set of n-levels coefficients.
        # Note: values are indexed according to model_level_number,
        # I.E. sigma, delta = PV[i], PV[NV/2+i] : where i=1..n_levels
        n_coeffs = n_levels + 1
        coeffs_array = np.zeros(n_coeffs * 2, dtype=np.float32)
        for n_lev, height, sigma in zip(model_levels,
                                        level_delta_coord.points,
                                        sigma_coord.points):
            # Record all the level coefficients coming from the 'full' cube.
            # Note: if some model levels are missing, we must still have the
            # coeffs at the correct index according to the model_level_number
            # value, i.e. at [level] and [NV // 2 + level].
            # Thus, we can *not* paste the values in a block: each one needs to
            # go in the index corresponding to its 'model_level_number' value.
            coeffs_array[n_lev] = height
            coeffs_array[n_coeffs + n_lev] = sigma
        pv_values = [float(el) for el in coeffs_array]
        # eccodes does not support writing numpy.int64, cast to python int
        gribapi.grib_set(grib, "NV", int(n_coeffs * 2))
        gribapi.grib_set_array(grib, "pv", pv_values)


def set_time_range(time_coord, grib):
    """
    Set the time range keys in the specified message
    based on the bounds of the provided time coordinate.

    """
    if len(time_coord.points) != 1:
        msg = 'Expected length one time coordinate, got {} points'
        raise ValueError(msg.format(len(time_coord.points)))

    if time_coord.nbounds != 2:
        msg = 'Expected time coordinate with two bounds, got {} bounds'
        raise ValueError(msg.format(time_coord.nbounds))

    # Set type to hours and convert period to this unit.
    gribapi.grib_set(grib, "indicatorOfUnitForTimeRange",
                     _TIME_RANGE_UNITS['hours'])
    hours_since_units = cf_units.Unit('hours since epoch',
                                      calendar=time_coord.units.calendar)
    start_hours, end_hours = time_coord.units.convert(time_coord.bounds[0],
                                                      hours_since_units)
    # Cast from np.float to Python int. The lengthOfTimeRange key is a
    # 4 byte integer so we cast to highlight truncation of any floating
    # point value. The grib_api will do the cast from float to int, but it
    # cannot handle numpy floats.
    time_range_in_hours = end_hours - start_hours
    integer_hours = int(time_range_in_hours)
    if integer_hours != time_range_in_hours:
        msg = 'Truncating floating point lengthOfTimeRange {} to ' \
              'integer value {}'
        warnings.warn(msg.format(time_range_in_hours, integer_hours))
    gribapi.grib_set(grib, "lengthOfTimeRange", integer_hours)


def set_time_increment(cell_method, grib):
    """
    Set the time increment keys in the specified message
    based on the provided cell method.

    """
    # Type of time increment, e.g incrementing forecast period, incrementing
    # forecast reference time, etc. Set to missing, but we could use the
    # cell method coord to infer a value (see code table 4.11).
    gribapi.grib_set(grib, "typeOfTimeIncrement", 255)

    # Default values for the time increment value and units type.
    inc = 0
    units_type = 255
    # Attempt to determine time increment from cell method intervals string.
    intervals = cell_method.intervals
    if intervals is not None and len(intervals) == 1:
        interval, = intervals
        try:
            inc, units = interval.split()
            inc = float(inc)
            if units in ('hr', 'hour', 'hours'):
                units_type = _TIME_RANGE_UNITS['hours']
            else:
                raise ValueError('Unable to parse units of interval')
        except ValueError:
            # Problem interpreting the interval string.
            inc = 0
            units_type = 255
        else:
            # Cast to int as timeIncrement key is a 4 byte integer.
            integer_inc = int(inc)
            if integer_inc != inc:
                warnings.warn('Truncating floating point timeIncrement {} to '
                              'integer value {}'.format(inc, integer_inc))
            inc = integer_inc

    gribapi.grib_set(grib, "indicatorOfUnitForTimeIncrement", units_type)
    gribapi.grib_set(grib, "timeIncrement", inc)


def _cube_is_time_statistic(cube):
    """
    Test whether we can identify this cube as a statistic over time.

    We need to know whether our cube represents a time statistic. This is
    almost always captured in the cell methods. The exception is when a
    percentage statistic has been calculated (i.e. for PDT10). This is
    captured in a `percentage_over_time` scalar coord, which must be handled
    here too.

    """
    result = False
    stat_coord_name = 'percentile_over_time'
    cube_coord_names = [coord.name() for coord in cube.coords()]

    # Check our cube for time statistic indicators.
    has_percentile_statistic = stat_coord_name in cube_coord_names
    has_cell_methods = cube.cell_methods

    # Determine whether we have a time statistic.
    if has_percentile_statistic:
        result = True
    elif has_cell_methods:
        # Define accepted time names, including from coord_categorisations.
        recognised_time_names = ['time', 'year', 'month', 'day', 'weekday',
                                 'season']
        latest_coordnames = cube.cell_methods[-1].coord_names
        if len(latest_coordnames) != 1:
            result = False
        else:
            coord_name = latest_coordnames[0]
            result = coord_name in recognised_time_names
    else:
        result = False

    return result


def _spatial_statistic(cube):
    """
    Gather and return the spatial statistic of the cube if it is present.

    """
    spatial_cell_methods = [
        cell_method for cell_method in cube.cell_methods if 'area' in
        cell_method.coord_names]

    if len(spatial_cell_methods) > 1:
        raise ValueError("Cannot handle multiple 'area' cell methods")
    elif len(spatial_cell_methods[0].coord_names) > 1:
        raise ValueError("Cannot handle multiple coordinate names in "
                         "the spatial processing related cell method. "
                         "Expected ('area',), got {!r}".format
                         (spatial_cell_methods[0].coord_names))

    return spatial_cell_methods


def statistical_method_code(cell_method_name):
    """
    Decode cell_method string as statistic code integer.
    """
    statistic_code = _STATISTIC_TYPE_NAMES.get(cell_method_name, None)
    if statistic_code is None:
        msg = ('Product definition section 4 contains an unsupported '
               'statistical process type [{}] ')
        raise TranslationError(msg.format(statistic_code))

    return statistic_code


def get_spatial_process_code(spatial_processing_type):
    """
    Decode spatial_processing_type string as spatial process code integer.
    """
    spatial_processing_code = None

    for code, interp_params in _SPATIAL_PROCESSING_TYPES.items():
        if interp_params.interpolation_type == spatial_processing_type:
            spatial_processing_code = code
            break

    if spatial_processing_code is None:
        msg = ('Product definition section 4 contains an unsupported '
               'spatial processing or interpolation type: {} ')
        raise TranslationError(msg.format(spatial_processing_type))

    return spatial_processing_code


def set_ensemble(cube, grib):
    """
    Set keys in the provided grib based message relating to ensemble
    information.

    """
    if not (cube.coords('realization') and
            len(cube.coord('realization').points) == 1):
        raise ValueError("A cube 'realization' coordinate with one "
                         "point is required, but not present")
    gribapi.grib_set(grib, "perturbationNumber",
                     int(cube.coord('realization').points[0]))
    # no encoding at present in iris-grib, set to missing
    gribapi.grib_set(grib, "numberOfForecastsInEnsemble", 255)
    gribapi.grib_set(grib, "typeOfEnsembleForecast", 255)


def product_definition_template_common(cube, grib, full3d_cube=None):
    """
    Set keys within the provided grib message that are common across
    all of the supported product definition templates.

    """
    set_discipline_and_parameter(cube, grib)

    # Various missing values.
    gribapi.grib_set(grib, "typeOfGeneratingProcess", 255)
    gribapi.grib_set(grib, "backgroundProcess", 255)
    gribapi.grib_set(grib, "generatingProcessIdentifier", 255)

    # Generic time handling.
    set_forecast_time(cube, grib)

    # Handle vertical coords.
    set_fixed_surfaces(cube, grib, full3d_cube)


def product_definition_template_0(cube, grib, full3d_cube=None):
    """
    Set keys within the provided grib message based on Product
    Definition Template 4.0.

    Template 4.0 is used to represent an analysis or forecast at
    a horizontal level at a point in time.

    """
    gribapi.grib_set_long(grib, "productDefinitionTemplateNumber", 0)
    product_definition_template_common(cube, grib, full3d_cube)


def product_definition_template_1(cube, grib, full3d_cube=None):
    """
    Set keys within the provided grib message based on Product
    Definition Template 4.1.

    Template 4.1 is used to represent an individual ensemble forecast, control
    and perturbed, at a horizontal level or in a horizontal layer at a point
    in time.

    """
    gribapi.grib_set(grib, "productDefinitionTemplateNumber", 1)
    product_definition_template_common(cube, grib, full3d_cube)
    set_ensemble(cube, grib)


def product_definition_template_8(cube, grib, full3d_cube=None):
    """
    Set keys within the provided grib message based on Product
    Definition Template 4.8.

    Template 4.8 is used to represent an aggregation over a time
    interval.

    """
    gribapi.grib_set(grib, "productDefinitionTemplateNumber", 8)
    _product_definition_template_8_10_and_11(cube, grib)


def product_definition_template_10(cube, grib, full3d_cube=None):
    """
    Set keys within the provided grib message based on Product Definition
    Template 4.10.

    Template 4.10 is used to represent a percentile forecast over a time
    interval.

    """
    gribapi.grib_set(grib, "productDefinitionTemplateNumber", 10)
    if not (cube.coords('percentile_over_time') and
            len(cube.coord('percentile_over_time').points) == 1):
        raise ValueError("A cube 'percentile_over_time' coordinate with one "
                         "point is required, but not present.")
    gribapi.grib_set(grib, "percentileValue",
                     int(cube.coord('percentile_over_time').points[0]))
    _product_definition_template_8_10_and_11(cube, grib)


def product_definition_template_11(cube, grib, full3d_cube=None):
    """
    Set keys within the provided grib message based on Product
    Definition Template 4.11.

    Template 4.11 is used to represent an aggregation over a time
    interval for an ensemble member.

    """
    gribapi.grib_set(grib, "productDefinitionTemplateNumber", 11)
    set_ensemble(cube, grib)
    _product_definition_template_8_10_and_11(cube, grib)


def _product_definition_template_8_10_and_11(cube, grib, full3d_cube=None):
    """
    Set keys within the provided grib message based on common aspects of
    Product Definition Templates 4.8 and 4.11.

    Templates 4.8 and 4.11 are used to represent aggregations over a time
    interval.

    """
    product_definition_template_common(cube, grib, full3d_cube)

    # Check for time coordinate.
    time_coord = cube.coord('time')

    if len(time_coord.points) != 1:
        msg = 'Expected length one time coordinate, got {} points'
        raise ValueError(msg.format(time_coord.points))

    if time_coord.nbounds != 2:
        msg = 'Expected time coordinate with two bounds, got {} bounds'
        raise ValueError(msg.format(time_coord.nbounds))

    # Extract the datetime-like object corresponding to the end of
    # the overall processing interval.
    end = time_coord.units.num2date(time_coord.bounds[0, -1])

    # Set the associated keys for the end of the interval (octets 35-41
    # in section 4).
    gribapi.grib_set(grib, "yearOfEndOfOverallTimeInterval", end.year)
    gribapi.grib_set(grib, "monthOfEndOfOverallTimeInterval", end.month)
    gribapi.grib_set(grib, "dayOfEndOfOverallTimeInterval", end.day)
    gribapi.grib_set(grib, "hourOfEndOfOverallTimeInterval", end.hour)
    gribapi.grib_set(grib, "minuteOfEndOfOverallTimeInterval", end.minute)
    gribapi.grib_set(grib, "secondOfEndOfOverallTimeInterval", end.second)

    # Only one time range specification. If there were a series of aggregations
    # (e.g. the mean of an accumulation) one might set this to a higher value,
    # but we currently only handle a single time related cell method.
    gribapi.grib_set(grib, "numberOfTimeRange", 1)
    gribapi.grib_set(grib, "numberOfMissingInStatisticalProcess", 0)

    # Period over which statistical processing is performed.
    set_time_range(time_coord, grib)

    # Check that there is one and only one cell method related to the
    # time coord.
    if cube.cell_methods:
        time_cell_methods = [
            cell_method for cell_method in cube.cell_methods if 'time' in
            cell_method.coord_names]
        if not time_cell_methods:
            raise ValueError("Expected a cell method with a coordinate name "
                             "of 'time'")
        if len(time_cell_methods) > 1:
            raise ValueError("Cannot handle multiple 'time' cell methods")
        cell_method, = time_cell_methods

        if len(cell_method.coord_names) > 1:
            raise ValueError("Cannot handle multiple coordinate names in "
                             "the time related cell method. Expected "
                             "('time',), got {!r}".format(
                                 cell_method.coord_names))

        # Type of statistical process (see code table 4.10)
        statistic_type = _STATISTIC_TYPE_NAMES.get(cell_method.method, 255)
        gribapi.grib_set(grib, "typeOfStatisticalProcessing", statistic_type)

        # Time increment i.e. interval of cell method (if any)
        set_time_increment(cell_method, grib)


def product_definition_template_15(cube, grib, full3d_cube=None):
    """
    Set keys within the provided grib message based on Product
    Definition Template 4.15.

    Template 4.15 represents the type of spatial processing used to
    arrive at the given data value from the source data.

    """
    # Encode type of spatial processing (see code table 4.15)
    spatial_processing_type = cube.attributes['spatial_processing_type']
    spatial_processing = get_spatial_process_code(spatial_processing_type)

    # Encode statistical process and number of points
    # (see template definition 4.15)
    statistical_process = _SPATIAL_PROCESSING_TYPES[spatial_processing][1]
    number_of_points = _SPATIAL_PROCESSING_TYPES[spatial_processing][2]

    # Only a limited number of spatial processing types are supported.
    if spatial_processing not in _SPATIAL_PROCESSING_TYPES.keys():
        msg = ('Cannot save Product Definition Type 4.15 with spatial '
               'processing type {}'.format(spatial_processing))
        raise ValueError(msg)

    if statistical_process is not None:
        # Check the cube for statistical cell methods over area
        spatial_stats = _spatial_statistic(cube)
        # Identify the statistical method (e.g. 'mean') and encode it.
        if len(spatial_stats) > 0:
            cell_method_name = spatial_stats[0].method
            statistical_process = statistical_method_code(cell_method_name)
        else:
            raise ValueError("Could not find a suitable cell_method to save "
                             "as a spatial statistical process.")

    # Set GRIB messages
    gribapi.grib_set(grib, "productDefinitionTemplateNumber", 15)
    product_definition_template_common(cube, grib, full3d_cube)
    gribapi.grib_set(grib, "spatialProcessing", spatial_processing)
    if number_of_points is not None:
        gribapi.grib_set(grib, "numberOfPointsUsed", number_of_points)
    if statistical_process is not None:
        gribapi.grib_set(grib, "statisticalProcess", statistical_process)


def product_definition_template_40(cube, grib, full3d_cube=None):
    """
    Set keys within the provided grib message based on Product
    Definition Template 4.40.

    Template 4.40 is used to represent an analysis or forecast at a horizontal
    level or in a horizontal layer at a point in time for atmospheric chemical
    constituents.

    """
    gribapi.grib_set(grib, "productDefinitionTemplateNumber", 40)
    product_definition_template_common(cube, grib)
    constituent_type = cube.attributes['WMO_constituent_type']
    gribapi.grib_set(grib, "constituentType", constituent_type)


def product_definition_section(cube, grib, full3d_cube=None):
    """
    Set keys within the product definition section of the provided
    grib message based on the properties of the cube.

    """
    if not cube.coord("time").has_bounds():
        if cube.coords('realization'):
            # ensemble forecast (template 4.1)
            pdt = product_definition_template_1(cube, grib, full3d_cube)
        elif 'WMO_constituent_type' in cube.attributes:
            # forecast for atmospheric chemical constiuent (template 4.40)
            product_definition_template_40(cube, grib, full3d_cube)
        elif 'spatial_processing_type' in cube.attributes:
            # spatial process (template 4.15)
            product_definition_template_15(cube, grib, full3d_cube)
        else:
            # forecast (template 4.0)
            product_definition_template_0(cube, grib, full3d_cube)
    elif _cube_is_time_statistic(cube):
        if cube.coords('realization'):
            # time processed (template 4.11)
            pdt = product_definition_template_11
        elif cube.coords('percentile_over_time'):
            # time processed as percentile (template 4.10)
            pdt = product_definition_template_10
        else:
            # time processed (template 4.8)
            pdt = product_definition_template_8
        try:
            pdt(cube, grib, full3d_cube)
        except ValueError as e:
            raise ValueError('Saving to GRIB2 failed: the cube is not suitable'
                             ' for saving as a time processed statistic GRIB'
                             ' message. {}'.format(e))
    else:
        # Don't know how to handle this kind of data
        msg = 'A suitable product template could not be deduced'
        raise TranslationError(msg)


###############################################################################
#
# Data Representation Section 5
#
###############################################################################

def data_section(cube, grib):
    # Masked data?
    if ma.isMaskedArray(cube.data):
        if not np.isnan(cube.data.fill_value):
            # Use the data's fill value.
            fill_value = float(cube.data.fill_value)
        else:
            # We can't use the cube's fill value if it's NaN,
            # the GRIB API doesn't like it.
            # Calculate an MDI outside the data range.
            min, max = cube.data.min(), cube.data.max()
            fill_value = min - (max - min) * 0.1
        # Prepare the unmasked data array, using fill_value as the MDI.
        data = cube.data.filled(fill_value)
    else:
        fill_value = None
        data = cube.data

    # units scaling
    grib2_info = gptx.cf_phenom_to_grib2_info(cube.standard_name,
                                              cube.long_name)
    if grib2_info is None:
        # for now, just allow this
        warnings.warn('Unable to determine Grib2 parameter code for cube.\n'
                      'Message data may not be correctly scaled.')
    else:
        if cube.units != grib2_info.units:
            data = cube.units.convert(data, grib2_info.units)
            if fill_value is not None:
                fill_value = cube.units.convert(fill_value, grib2_info.units)

    if fill_value is None:
        # Disable missing values in the grib message.
        gribapi.grib_set(grib, "bitmapPresent", 0)
    else:
        # Enable missing values in the grib message.
        gribapi.grib_set(grib, "bitmapPresent", 1)
        gribapi.grib_set_double(grib, "missingValue", fill_value)

    gribapi.grib_set_double_array(grib, "values", data.flatten())

    # todo: check packing accuracy?
#    print("packingError", gribapi.getb_get_double(grib, "packingError"))


###############################################################################

def gribbability_check(cube):
    "We always need the following things for grib saving."

    # GeogCS exists?
    cs0 = cube.coord(dimensions=[0]).coord_system
    cs1 = cube.coord(dimensions=[1]).coord_system
    if cs0 is None or cs1 is None:
        raise TranslationError("CoordSystem not present")
    if cs0 != cs1:
        raise TranslationError("Inconsistent CoordSystems")

    # Time period exists?
    if not cube.coords("time"):
        raise TranslationError("time coord not found")


def run(slice2d_cube, grib, full3d_cube):
    """
    Set the keys of the grib message based on the contents of the slice2d_cube.

    Args:

    * slice2d_cube:
        A :class:`iris.slice2d_cube.Cube` representing a 2d field.

    * grib_message_id:
        ID of a grib message in memory. This is typically the return value of
        :func:`gribapi.grib_new_from_samples`.

    * full3d_cube:
        A :class:`iris.slice2d_cube.Cube` representing the entire save cube.
        This is required to write data with hybrid vertical coordinates, as
        the GRIB2 spec records hybrid coefficients for *all* the levels in
        every message.

    """
    gribbability_check(slice2d_cube)

    # Section 1 - Identification Section.
    identification(slice2d_cube, grib)

    # Section 3 - Grid Definition Section (Grid Definition Template)
    grid_definition_section(slice2d_cube, grib)

    # Section 4 - Product Definition Section (Product Definition Template)
    product_definition_section(slice2d_cube, grib, full3d_cube)

    # Section 5 - Data Representation Section (Data Representation Template)
    data_section(slice2d_cube, grib)



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for :func:`iris_grib._save_rules.set_fixed_surfaces`.

"""

# Import iris_grib.tests first so that some things can be initialised before
# importing anything else.
import iris_grib.tests as tests

from unittest import mock

import gribapi
import numpy as np

import iris.cube
import iris.coords

from iris_grib._save_rules import set_fixed_surfaces


class Test(tests.IrisGribTest):
    def test_bounded_altitude_feet(self):
        cube = iris.cube.Cube([0])
        cube.add_aux_coord(iris.coords.AuxCoord(
            1500.0, long_name='altitude', units='ft',
            bounds=np.array([1000.0, 2000.0])))
        grib = gribapi.grib_new_from_samples("GRIB2")
        set_fixed_surfaces(cube, grib)
        self.assertEqual(
            gribapi.grib_get_double(grib, "scaledValueOfFirstFixedSurface"),
            305.0)  # precise ~304.8
        self.assertEqual(
            gribapi.grib_get_double(grib, "scaledValueOfSecondFixedSurface"),
            610.0)  # precise ~609.6
        self.assertEqual(
            gribapi.grib_get_long(grib, "typeOfFirstFixedSurface"),
            102)
        self.assertEqual(
            gribapi.grib_get_long(grib, "typeOfSecondFixedSurface"),
            102)

    def test_theta_level(self):
        cube = iris.cube.Cube([0])
        cube.add_aux_coord(iris.coords.AuxCoord(
            230.0, standard_name='air_potential_temperature',
            units='K', attributes={'positive': 'up'},
            bounds=np.array([220.0, 240.0])))
        grib = gribapi.grib_new_from_samples("GRIB2")
        set_fixed_surfaces(cube, grib)
        self.assertEqual(
            gribapi.grib_get_double(grib, "scaledValueOfFirstFixedSurface"),
            220.0)
        self.assertEqual(
            gribapi.grib_get_double(grib, "scaledValueOfSecondFixedSurface"),
            240.0)
        self.assertEqual(
            gribapi.grib_get_long(grib, "typeOfFirstFixedSurface"),
            107)
        self.assertEqual(
            gribapi.grib_get_long(grib, "typeOfSecondFixedSurface"),
            107)

    def test_depth(self):
        cube = iris.cube.Cube([0])
        cube.add_aux_coord(iris.coords.AuxCoord(
            1, long_name='depth', units='m',
            bounds=np.array([0., 2]), attributes={'positive': 'down'}))
        grib = gribapi.grib_new_from_samples("GRIB2")
        set_fixed_surfaces(cube, grib)
        self.assertEqual(
            gribapi.grib_get_double(grib, "scaledValueOfFirstFixedSurface"),
            0.)
        self.assertEqual(
            gribapi.grib_get_double(grib, "scaledValueOfSecondFixedSurface"),
            2)
        self.assertEqual(
            gribapi.grib_get_long(grib, "typeOfFirstFixedSurface"),
            106)
        self.assertEqual(
            gribapi.grib_get_long(grib, "typeOfSecondFixedSurface"),
            106)

    @mock.patch.object(gribapi, "grib_set")
    def test_altitude_point(self, mock_set):
        grib = None
        cube = iris.cube.Cube([1, 2, 3, 4, 5])
        cube.add_aux_coord(
            iris.coords.AuxCoord([12345], "altitude", units="m")
        )

        set_fixed_surfaces(cube, grib)

        mock_set.assert_any_call(grib, "typeOfFirstFixedSurface", 102)
        mock_set.assert_any_call(grib, "scaleFactorOfFirstFixedSurface", 0)
        mock_set.assert_any_call(grib, "scaledValueOfFirstFixedSurface",
                                 12345)
        mock_set.assert_any_call(grib, "typeOfSecondFixedSurface", -1)
        mock_set.assert_any_call(grib, "scaleFactorOfSecondFixedSurface",
                                 255)
        mock_set.assert_any_call(grib, "scaledValueOfSecondFixedSurface",
                                 -1)

    @mock.patch.object(gribapi, "grib_set")
    def test_height_point(self, mock_set):
        grib = None
        cube = iris.cube.Cube([1, 2, 3, 4, 5])
        cube.add_aux_coord(iris.coords.AuxCoord([12345], "height", units="m"))

        set_fixed_surfaces(cube, grib)

        mock_set.assert_any_call(grib, "typeOfFirstFixedSurface", 103)
        mock_set.assert_any_call(grib, "scaleFactorOfFirstFixedSurface", 0)
        mock_set.assert_any_call(grib, "scaledValueOfFirstFixedSurface", 12345)
        mock_set.assert_any_call(grib, "typeOfSecondFixedSurface", -1)
        mock_set.assert_any_call(grib, "scaleFactorOfSecondFixedSurface", 255)
        mock_set.assert_any_call(grib, "scaledValueOfSecondFixedSurface", -1)

    @mock.patch.object(gribapi, "grib_set")
    def test_no_vertical(self, mock_set):
        grib = None
        cube = iris.cube.Cube([1, 2, 3, 4, 5])
        set_fixed_surfaces(cube, grib)
        mock_set.assert_any_call(grib, "typeOfFirstFixedSurface", 1)
        mock_set.assert_any_call(grib, "scaleFactorOfFirstFixedSurface", 0)
        mock_set.assert_any_call(grib, "scaledValueOfFirstFixedSurface", 0)
        mock_set.assert_any_call(grib, "typeOfSecondFixedSurface", -1)
        mock_set.assert_any_call(grib, "scaleFactorOfSecondFixedSurface", 255)
        mock_set.assert_any_call(grib, "scaledValueOfSecondFixedSurface", -1)


if __name__ == "__main__":
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Tests for function :func:`iris_grib._load_convert.data_cutoff`.

"""

# import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests

from unittest import mock

from iris_grib._load_convert import _MDI as MDI
from iris_grib._load_convert import data_cutoff


class TestDataCutoff(tests.IrisGribTest):
    def _check(self, hours, minutes, request_warning, expect_warning=False):
        # Setup the environment.
        patch_target = 'iris_grib._load_convert.options'
        with mock.patch(patch_target) as options:
            options.warn_on_unsupported = request_warning
            with mock.patch('warnings.warn') as warn:
                # The call being tested.
                data_cutoff(hours, minutes)
        # Check the result.
        if expect_warning:
            self.assertEqual(len(warn.mock_calls), 1)
            args, kwargs = warn.call_args
            self.assertIn('data cutoff', args[0])
        else:
            self.assertEqual(len(warn.mock_calls), 0)

    def test_neither(self):
        self._check(MDI, MDI, False)

    def test_hours(self):
        self._check(3, MDI, False)

    def test_minutes(self):
        self._check(MDI, 20, False)

    def test_hours_and_minutes(self):
        self._check(30, 40, False)

    def test_neither_warning(self):
        self._check(MDI, MDI, True, False)

    def test_hours_warning(self):
        self._check(3, MDI, True, True)

    def test_minutes_warning(self):
        self._check(MDI, 20, True, True)

    def test_hours_and_minutes_warning(self):
        self._check(30, 40, True, True)


if __name__ == '__main__':
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for `iris_grib._load_convert._calculate_increment`.

"""

# Import iris_grib.tests first so that some things can be initialised before
# importing anything else.
import iris_grib.tests as tests

from iris_grib._load_convert import _calculate_increment


class Test(tests.IrisGribTest):
    def test_negative(self):
        result = _calculate_increment(-15, -5, 10)
        self.assertEqual(result, 1)

    def test_positive(self):
        result = _calculate_increment(-5, 5, 10)
        self.assertEqual(result, 1)

    def test_with_mod(self):
        result = _calculate_increment(355, 5, 10, 360)
        self.assertEqual(result, 1)


if __name__ == '__main__':
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for
:func:`iris_grib._load_convert.grid_definition_template_12`.

"""

# import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests

import numpy as np

import iris.coord_systems
import iris.coords
import iris.exceptions

from iris_grib.tests.unit.load_convert import empty_metadata
from iris_grib._load_convert import _MDI as MDI

from iris_grib._load_convert import grid_definition_template_12


class Test(tests.IrisGribTest):
    def section_3(self):
        section = {
            'shapeOfTheEarth': 7,
            'scaleFactorOfRadiusOfSphericalEarth': MDI,
            'scaledValueOfRadiusOfSphericalEarth': MDI,
            'scaleFactorOfEarthMajorAxis': 3,
            'scaledValueOfEarthMajorAxis': 6377563396,
            'scaleFactorOfEarthMinorAxis': 3,
            'scaledValueOfEarthMinorAxis': 6356256909,
            'Ni': 4,
            'Nj': 3,
            'latitudeOfReferencePoint': 49000000,
            'longitudeOfReferencePoint': -2000000,
            'resolutionAndComponentFlags': 0,
            'scaleFactorAtReferencePoint': 0.9996012717,
            'XR': 40000000,
            'YR': -10000000,
            'scanningMode': 64,
            'Di': 200000,
            'Dj': 100000,
            'X1': 29300000,
            'Y1': 9200000,
            'X2': 29900000,
            'Y2': 9400000
        }
        return section

    def expected(self, y_dim, x_dim):
        # Prepare the expectation.
        expected = empty_metadata()
        ellipsoid = iris.coord_systems.GeogCS(6377563.396, 6356256.909)
        cs = iris.coord_systems.TransverseMercator(49, -2, 400000, -100000,
                                                   0.9996012717, ellipsoid)
        nx = 4
        x_origin = 293000
        dx = 2000
        x = iris.coords.DimCoord(np.arange(nx) * dx + x_origin,
                                 'projection_x_coordinate', units='m',
                                 coord_system=cs)
        ny = 3
        y_origin = 92000
        dy = 1000
        y = iris.coords.DimCoord(np.arange(ny) * dy + y_origin,
                                 'projection_y_coordinate', units='m',
                                 coord_system=cs)
        expected['dim_coords_and_dims'].append((y, y_dim))
        expected['dim_coords_and_dims'].append((x, x_dim))
        return expected

    def test(self):
        section = self.section_3()
        metadata = empty_metadata()
        grid_definition_template_12(section, metadata)
        expected = self.expected(0, 1)
        self.assertEqual(metadata, expected)

    def test_spherical(self):
        section = self.section_3()
        section['shapeOfTheEarth'] = 0
        metadata = empty_metadata()
        grid_definition_template_12(section, metadata)
        expected = self.expected(0, 1)
        cs = expected['dim_coords_and_dims'][0][0].coord_system
        cs.ellipsoid = iris.coord_systems.GeogCS(6367470)
        self.assertEqual(metadata, expected)

    def test_negative_x(self):
        section = self.section_3()
        section['scanningMode'] = 0b11000000
        metadata = empty_metadata()
        with self.assertRaisesRegex(iris.exceptions.TranslationError,
                                    '-x scanning'):
            grid_definition_template_12(section, metadata)

    def test_negative_y(self):
        section = self.section_3()
        section['scanningMode'] = 0b00000000
        metadata = empty_metadata()
        with self.assertRaisesRegex(iris.exceptions.TranslationError,
                                    '-y scanning'):
            grid_definition_template_12(section, metadata)

    def test_transposed(self):
        section = self.section_3()
        section['scanningMode'] = 0b01100000
        metadata = empty_metadata()
        grid_definition_template_12(section, metadata)
        expected = self.expected(1, 0)
        self.assertEqual(metadata, expected)

    def test_di_tolerance(self):
        # Even though Ni * Di doesn't exactly match X1 to X2 it should
        # be close enough to allow the translation.
        section = self.section_3()
        section['X2'] += 1
        metadata = empty_metadata()
        grid_definition_template_12(section, metadata)
        expected = self.expected(0, 1)
        x = expected['dim_coords_and_dims'][1][0]
        x.points = np.linspace(293000, 299000.01, 4)
        self.assertEqual(metadata, expected)

    def test_incompatible_grid_extent(self):
        section = self.section_3()
        section['X2'] += 100
        metadata = empty_metadata()
        with self.assertRaisesRegex(iris.exceptions.TranslationError,
                                    'grid'):
            grid_definition_template_12(section, metadata)

    def test_scale_workaround(self):
        section = self.section_3()
        section['scaleFactorAtReferencePoint'] = 1065346526
        metadata = empty_metadata()
        grid_definition_template_12(section, metadata)
        expected = self.expected(0, 1)
        # A float32 can't hold exactly the same value.
        cs = expected['dim_coords_and_dims'][0][0].coord_system
        cs.scale_factor_at_central_meridian = 0.9996012449264526
        self.assertEqual(metadata, expected)


if __name__ == '__main__':
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for :func:`iris_grib._save_rules.data_section`.

"""

# import iris_grib.tests first so that some things can be initialised before
# importing anything else
import iris_grib.tests as tests

import numpy as np

import iris
from iris.cube import Cube

from iris_grib import save_pairs_from_cube, save_messages, GribMessage


@tests.skip_grib_data
class TestSaveHybridHeight(tests.IrisGribTest):
    def setUp(self):
        reference_data_filepath = self.get_testdata_path('hybrid_height.nc')
        if (hasattr(iris.FUTURE, 'netcdf_promote') and
                not iris.FUTURE.netcdf_promote):
            iris.FUTURE.netcdf_promote = True
        data_cube = iris.load_cube(reference_data_filepath,
                                   'air_potential_temperature')
        # Use only 3 (non-contiguous) levels, and a single timestep.
        data_cube = data_cube[0, :6:2]
        self.test_hh_data_cube = data_cube

    def test_save(self):
        # Get save-pairs for the test data.
        save_pairs = save_pairs_from_cube(self.test_hh_data_cube)

        # Check there are 3 of them (and nothing failed !)
        save_pairs = list(save_pairs)
        self.assertEqual(len(save_pairs), 3)

        # Get a list of just the messages.
        msgs = [pair[1] for pair in save_pairs]

        # Save the messages to a temporary file.
        with self.temp_filename() as temp_path:
            save_messages(msgs, temp_path, append=True)

            # Read back as GribMessage-s.
            msgs = list(GribMessage.messages_from_filename(temp_path))

            # Check 3 messages were saved.
            self.assertEqual(len(msgs), 3)

            # Check that the PV vector (same in all messages) is as expected.
            # Note: gaps here are because we took model levels = (1, 3, 5).
            self.assertArrayAllClose(
                msgs[0].sections[4]['pv'],
                [0, 5., 0, 45., 0, 111.667, 0, 0.999, 0, 0.995, 0, 0.987],
                atol=0.0015)

            # Check message #2-of-3 has the correctly encoded hybrid height.
            msg = msgs[1]
            #  first surface type = 118  (i.e. hybrid height).
            self.assertEqual(
                msg.sections[4]['typeOfFirstFixedSurface'],
                118)
            #  first surface scaling = 0.
            self.assertEqual(
                msg.sections[4]['scaleFactorOfFirstFixedSurface'],
                0)
            #  first surface value = 3  -- i.e. #2 of (1, 3, 5).
            self.assertEqual(
                msg.sections[4]['scaledValueOfFirstFixedSurface'],
                3)
            #  second surface type = "NONE"  -- i.e. unbounded level.
            self.assertEqual(
                msg.sections[4]['typeOfSecondFixedSurface'],
                255)


@tests.skip_grib_data
class TestSaveHybridPressure(tests.IrisGribTest):
    def setUp(self):
        reference_data_filepath = self.get_testdata_path(
            'hybrid_pressure.nc')
        data_cube = iris.load_cube(reference_data_filepath,
                                   'air_temperature')
        self.test_hp_data_cube = data_cube

    def test_save(self):
        # Get save-pairs for the test data.
        save_pairs = save_pairs_from_cube(self.test_hp_data_cube)

        # Check there are 3 of them (and nothing failed !)
        save_pairs = list(save_pairs)
        self.assertEqual(len(save_pairs), 3)

        # Get a list of just the messages.
        msgs = [pair[1] for pair in save_pairs]

        # Save the messages to a temporary file.
        with self.temp_filename() as temp_path:
            save_messages(msgs, temp_path, append=True)

            # Read back as GribMessage-s.
            msgs = list(GribMessage.messages_from_filename(temp_path))

            # Check 3 messages were saved.
            self.assertEqual(len(msgs), 3)

            # Check that the PV vector (same in all messages) is as expected.
            # Note: HUGE gaps here because we took model levels = (1, 51, 91).
            self.assertEqual(msgs[0].sections[4]['NV'], 184)
            pv_expected = np.zeros(184, dtype=np.float64)
            pv_expected[[1, 51, 91]] = [0., 18191.03, 0.003]
            pv_expected[[93, 143, 183]] = [0., 0.036, 0.998]
            self.assertArrayAllClose(
                msgs[0].sections[4]['pv'], pv_expected, atol=0.001)

            # Check message #2-of-3 has the correctly encoded hybrid pressure.
            msg = msgs[1]
            #  first surface type = 119  (i.e. hybrid pressure).
            self.assertEqual(
                msg.sections[4]['typeOfFirstFixedSurface'],
                119)
            #  first surface scaling = 0.
            self.assertEqual(
                msg.sections[4]['scaleFactorOfFirstFixedSurface'],
                0)
            #  first surface value = 3  -- i.e. #2 of (1, 3, 5).
            self.assertEqual(
                msg.sections[4]['scaledValueOfFirstFixedSurface'],
                51)
            #  second surface type = "NONE"  -- i.e. unbounded level.
            self.assertEqual(
                msg.sections[4]['typeOfSecondFixedSurface'],
                255)


if __name__ == '__main__':
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for `iris_grib.grib_save_rules.set_discipline_and_parameter`."""
# Import iris_grib.tests first so that some things can be initialised before
# importing anything else.
import iris_grib.tests as tests
from unittest import mock

from iris.cube import Cube

from iris_grib.grib_phenom_translation import GRIBCode

from iris_grib._save_rules import set_discipline_and_parameter


class TestPhenomenonCoding(tests.IrisGribTest):
    def setUp(self):
        # A mock cube with empty phenomenon-specifying metadata.
        self.mock_cube = mock.Mock(
            spec=Cube,
            standard_name=None,
            long_name=None,
            attributes={})

    def _check_coding(self, cube, discipline, paramCategory, paramNumber):
        # Check that encoding 'cube' writes the expected phenomenon keys.
        grib_set_patch = self.patch(
            'iris_grib._save_rules.gribapi.grib_set')
        mock_message = mock.sentinel.grib2_message

        set_discipline_and_parameter(cube, mock_message)

        expected_calls = [
            mock.call(mock_message, "discipline", discipline),
            mock.call(mock_message, "parameterCategory", paramCategory),
            mock.call(mock_message, "parameterNumber", paramNumber)]

        self.assertEqual(grib_set_patch.call_args_list, expected_calls)

    def test_unknown_phenomenon(self):
        cube = self.mock_cube
        self._check_coding(cube, 255, 255, 255)

    def test_known_standard_name(self):
        cube = self.mock_cube
        cube.standard_name = 'sea_water_y_velocity'
        self._check_coding(cube, 10, 1, 3)  # as seen in _grib_cf_map.py

    def test_known_long_name(self):
        cube = self.mock_cube
        cube.long_name = 'cloud_mixing_ratio'
        self._check_coding(cube, 0, 1, 22)

    def test_gribcode_attribute_object(self):
        cube = self.mock_cube
        cube.attributes = {'GRIB_PARAM': GRIBCode(2, 7, 12, 99)}
        self._check_coding(cube, 7, 12, 99)

    def test_gribcode_attribute_string(self):
        cube = self.mock_cube
        cube.attributes = {'GRIB_PARAM': '2, 9, 33, 177'}
        self._check_coding(cube, 9, 33, 177)

    def test_gribcode_attribute_tuple(self):
        cube = self.mock_cube
        cube.attributes = {'GRIB_PARAM': (2, 33, 4, 12)}
        self._check_coding(cube, 33, 4, 12)

    def test_gribcode_attribute_not_edition_2(self):
        cube = self.mock_cube
        cube.attributes = {'GRIB_PARAM': GRIBCode(1, 7, 12, 99)}
        self._check_coding(cube, 255, 255, 255)

    def test_gribcode_attribute_overrides_phenomenon(self):
        cube = self.mock_cube
        cube.standard_name = 'sea_water_y_velocity'
        cube.attributes = {'GRIB_PARAM': '2, 9, 33, 177'}
        self._check_coding(cube, 9, 33, 177)


if __name__ == "__main__":
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Test function :func:`iris_grib._load_convert.ellipsoid_geometry.

"""

# import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests

from iris_grib._load_convert import ellipsoid_geometry


class Test(tests.IrisGribTest):
    def setUp(self):
        self.section = {'scaledValueOfEarthMajorAxis': 10,
                        'scaleFactorOfEarthMajorAxis': 1,
                        'scaledValueOfEarthMinorAxis': 100,
                        'scaleFactorOfEarthMinorAxis': 2,
                        'scaledValueOfRadiusOfSphericalEarth': 1000,
                        'scaleFactorOfRadiusOfSphericalEarth': 3}

    def test_geometry(self):
        result = ellipsoid_geometry(self.section)
        self.assertEqual(result, (1.0, 1.0, 1.0))


if __name__ == '__main__':
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for the `iris_grib.load_pairs_from_fields` function."""

import iris_grib.tests as tests

from iris_grib import load_pairs_from_fields
from iris_grib.message import GribMessage


class TestAsCubes(tests.IrisTest):
    def setUp(self):
        # Load from the test file.
        self.file_path = tests.get_data_path(
            ("GRIB", "time_processed", "time_bound.grib2")
        )

    def test_year_filter(self):
        msgs = GribMessage.messages_from_filename(self.file_path)
        chosen_messages = []
        for gmsg in msgs:
            if gmsg.sections[1]["year"] == 1998:
                chosen_messages.append(gmsg)
        cubes_msgs = list(load_pairs_from_fields(chosen_messages))
        self.assertEqual(len(cubes_msgs), 1)

    def test_year_filter_none(self):
        msgs = GribMessage.messages_from_filename(self.file_path)
        chosen_messages = []
        for gmsg in msgs:
            if gmsg.sections[1]["year"] == 1958:
                chosen_messages.append(gmsg)
        cubes_msgs = list(load_pairs_from_fields(chosen_messages))
        self.assertEqual(len(cubes_msgs), 0)

    def test_as_pairs(self):
        messages = GribMessage.messages_from_filename(self.file_path)
        cubes = []
        cube_msg_pairs = load_pairs_from_fields(messages)
        for cube, gmsg in cube_msg_pairs:
            if gmsg.sections[1]["year"] == 1998:
                cube.attributes["the year is"] = gmsg.sections[1]["year"]
                cubes.append(cube)
        self.assertEqual(len(cubes), 1)
        self.assertEqual(cubes[0].attributes["the year is"], 1998)


if __name__ == "__main__":
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for
:func:`iris_grib._load_convert.grid_definition_template_30`.
"""

# import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests

import cartopy.crs as ccrs
import numpy as np

import iris.coord_systems
import iris.coords

from iris_grib.tests.unit.load_convert import empty_metadata
from iris_grib._load_convert import _MDI as MDI

from iris_grib._load_convert import grid_definition_template_30


class Test(tests.IrisGribTest):

    def section_3(self):
        section = {
            'gridDefinitionTemplateNumber': 30,
            'shapeOfTheEarth': 0,
            'scaleFactorOfRadiusOfSphericalEarth': 0,
            'scaledValueOfRadiusOfSphericalEarth': 6367470,
            'scaleFactorOfEarthMajorAxis': 0,
            'scaledValueOfEarthMajorAxis': MDI,
            'scaleFactorOfEarthMinorAxis': 0,
            'scaledValueOfEarthMinorAxis': MDI,
            'Nx': 15,
            'Ny': 10,
            'longitudeOfFirstGridPoint': 239550000,
            'latitudeOfFirstGridPoint': 21641000,
            'resolutionAndComponentFlags': 0b00001000,
            'LaD': 60000000,
            'LoV': 262000000,
            'Dx': 320000000,
            'Dy': 320000000,
            'projectionCentreFlag': 0b00000000,
            'scanningMode': 0b01000000,
            'Latin1': 60000000,
            'Latin2': 30000000,
        }
        return section

    def expected(self, y_dim, x_dim):
        # Prepare the expectation.
        expected = empty_metadata()
        cs = iris.coord_systems.GeogCS(6367470)
        cs = iris.coord_systems.LambertConformal(
            central_lat=60.,
            central_lon=262.,
            false_easting=0,
            false_northing=0,
            secant_latitudes=(60., 30.),
            ellipsoid=iris.coord_systems.GeogCS(6367470))
        lon0 = 239.55
        lat0 = 21.641
        x0m, y0m = cs.as_cartopy_crs().transform_point(
            lon0, lat0, ccrs.Geodetic())
        dxm = dym = 320000.
        x_points = x0m + dxm * np.arange(15)
        y_points = y0m + dym * np.arange(10)
        x = iris.coords.DimCoord(x_points,
                                 standard_name='projection_x_coordinate',
                                 units='m',
                                 coord_system=cs,
                                 circular=False)
        y = iris.coords.DimCoord(y_points,
                                 standard_name='projection_y_coordinate',
                                 units='m',
                                 coord_system=cs)
        expected['dim_coords_and_dims'].append((y, y_dim))
        expected['dim_coords_and_dims'].append((x, x_dim))
        return expected

    def test(self):
        section = self.section_3()
        metadata = empty_metadata()
        grid_definition_template_30(section, metadata)
        expected = self.expected(0, 1)
        self.assertEqual(metadata, expected)


if __name__ == '__main__':
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for the `iris_grib.load_cubes` function."""

import iris_grib.tests as tests

from unittest import mock

import iris
from iris.fileformats.rules import Loader

import iris_grib
from iris_grib import load_cubes


class Test(tests.IrisGribTest):
    def test(self):
        generator = iris_grib._load_generate
        converter = iris_grib._load_convert.convert
        files = mock.sentinel.FILES
        callback = mock.sentinel.CALLBACK
        expected_result = mock.sentinel.RESULT
        with mock.patch('iris.fileformats.rules.load_cubes') as rules_load:
            rules_load.return_value = expected_result
            result = load_cubes(files, callback)
            kwargs = {}
            loader = Loader(generator, kwargs, converter)
            rules_load.assert_called_once_with(files, callback, loader)
            self.assertIs(result, expected_result)


@tests.skip_data
class Test_load_cubes(tests.IrisGribTest):

    def test_reduced_raw(self):
        # Loading a GRIB message defined on a reduced grid without
        # interpolating to a regular grid.
        gribfile = tests.get_data_path(
            ("GRIB", "reduced", "reduced_gg.grib2"))
        grib_generator = load_cubes(gribfile)
        self.assertCML(next(grib_generator))


if __name__ == "__main__":
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Test function :func:`iris_grib._load_convert.bitmap_section.`

"""

# import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests

from iris.exceptions import TranslationError

from iris_grib._load_convert import bitmap_section
from iris_grib.tests.unit import _make_test_message


class Test(tests.IrisGribTest):
    def test_bitmap_unsupported(self):
        # bitMapIndicator in range 1-254.
        # Note that bitMapIndicator = 1-253 and bitMapIndicator = 254 mean two
        # different things, but load_convert treats them identically.
        message = _make_test_message({6: {'bitMapIndicator': 100,
                                          'bitmap': None}})
        with self.assertRaisesRegex(TranslationError, 'unsupported bitmap'):
            bitmap_section(message.sections[6])


if __name__ == '__main__':
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for the `iris_grib.message.GribMessage` class.

"""

# Import iris_grib.tests first so that some things can be initialised before
# importing anything else.
import iris_grib.tests as tests

from abc import ABCMeta, abstractmethod
from unittest import mock

import numpy as np
import numpy.ma as ma

from iris._lazy_data import as_concrete_data, is_lazy_data
from iris.exceptions import TranslationError

from iris_grib.message import GribMessage
from iris_grib.tests.unit import _make_test_message


SECTION_6_NO_BITMAP = {'bitMapIndicator': 255, 'bitmap': None}


@tests.skip_data
class Test_messages_from_filename(tests.IrisGribTest):
    def test(self):
        filename = tests.get_data_path(('GRIB', '3_layer_viz',
                                        '3_layer.grib2'))
        messages = list(GribMessage.messages_from_filename(filename))
        self.assertEqual(len(messages), 3)

    def test_release_file(self):
        filename = tests.get_data_path(('GRIB', '3_layer_viz',
                                        '3_layer.grib2'))
        my_file = open(filename)

        import builtins
        self.patch('builtins.open', mock.Mock(return_value=my_file))

        messages = list(GribMessage.messages_from_filename(filename))
        self.assertFalse(my_file.closed)
        del messages
        self.assertTrue(my_file.closed)


class Test_sections(tests.IrisGribTest):
    def test(self):
        # Check that the `sections` attribute defers to the `sections`
        # attribute on the underlying _RawGribMessage.
        message = _make_test_message(mock.sentinel.SECTIONS)
        self.assertIs(message.sections, mock.sentinel.SECTIONS)


class Test_data__masked(tests.IrisGribTest):
    def setUp(self):
        self.bitmap = np.array([0, 0, 1, 1, 0, 0, 0, 1, 0, 1, 0, 1])
        self.shape = (3, 4)
        self._section_3 = {'sourceOfGridDefinition': 0,
                           'numberOfOctectsForNumberOfPoints': 0,
                           'interpretationOfNumberOfPoints': 0,
                           'gridDefinitionTemplateNumber': 0,
                           'scanningMode': 0,
                           'Nj': self.shape[0],
                           'Ni': self.shape[1]}

    def test_no_bitmap(self):
        values = np.arange(12)
        message = _make_test_message({3: self._section_3,
                                      6: SECTION_6_NO_BITMAP,
                                      7: {'codedValues': values}})
        result = as_concrete_data(message.data)
        expected = values.reshape(self.shape)
        self.assertEqual(result.shape, self.shape)
        self.assertArrayEqual(result, expected)

    def test_bitmap_present(self):
        # Test the behaviour where bitmap and codedValues shapes
        # are not equal.
        input_values = np.arange(5)
        output_values = np.array([-1, -1, 0, 1, -1, -1, -1, 2, -1, 3, -1, 4])
        message = _make_test_message({3: self._section_3,
                                      6: {'bitMapIndicator': 0,
                                          'bitmap': self.bitmap},
                                      7: {'codedValues': input_values}})
        result = as_concrete_data(message.data)
        expected = ma.masked_array(output_values,
                                   np.logical_not(self.bitmap))
        expected = expected.reshape(self.shape)
        self.assertMaskedArrayEqual(result, expected)

    def test_bitmap__shapes_mismatch(self):
        # Test the behaviour where bitmap and codedValues shapes do not match.
        # Too many or too few unmasked values in codedValues will cause this.
        values = np.arange(6)
        message = _make_test_message({3: self._section_3,
                                      6: {'bitMapIndicator': 0,
                                          'bitmap': self.bitmap},
                                      7: {'codedValues': values}})
        with self.assertRaisesRegex(TranslationError, 'do not match'):
            as_concrete_data(message.data)

    def test_bitmap__invalid_indicator(self):
        values = np.arange(12)
        message = _make_test_message({3: self._section_3,
                                      6: {'bitMapIndicator': 100,
                                          'bitmap': None},
                                      7: {'codedValues': values}})
        with self.assertRaisesRegex(TranslationError, 'unsupported bitmap'):
            as_concrete_data(message.data)


class Test_data__unsupported(tests.IrisGribTest):
    def test_unsupported_grid_definition(self):
        message = _make_test_message({3: {'sourceOfGridDefinition': 1},
                                      6: SECTION_6_NO_BITMAP})
        with self.assertRaisesRegex(TranslationError, 'source'):
            message.data

    def test_unsupported_quasi_regular__number_of_octets(self):
        message = _make_test_message(
            {3: {'sourceOfGridDefinition': 0,
                 'numberOfOctectsForNumberOfPoints': 1,
                 'gridDefinitionTemplateNumber': 0},
             6: SECTION_6_NO_BITMAP})
        with self.assertRaisesRegex(TranslationError, 'quasi-regular'):
            message.data

    def test_unsupported_quasi_regular__interpretation(self):
        message = _make_test_message(
            {3: {'sourceOfGridDefinition': 0,
                 'numberOfOctectsForNumberOfPoints': 0,
                 'interpretationOfNumberOfPoints': 1,
                 'gridDefinitionTemplateNumber': 0},
             6: SECTION_6_NO_BITMAP})
        with self.assertRaisesRegex(TranslationError, 'quasi-regular'):
            message.data

    def test_unsupported_template(self):
        message = _make_test_message(
            {3: {'sourceOfGridDefinition': 0,
                 'numberOfOctectsForNumberOfPoints': 0,
                 'interpretationOfNumberOfPoints': 0,
                 'gridDefinitionTemplateNumber': 2}})
        with self.assertRaisesRegex(TranslationError, 'template'):
            message.data


# Abstract, mix-in class for testing the `data` attribute for various
# grid definition templates.
class Mixin_data__grid_template(metaclass=ABCMeta):
    @abstractmethod
    def section_3(self, scanning_mode):
        raise NotImplementedError()

    def test_unsupported_scanning_mode(self):
        message = _make_test_message(
            {3: self.section_3(1),
             6: SECTION_6_NO_BITMAP})
        with self.assertRaisesRegex(TranslationError, 'scanning mode'):
            message.data

    def _test(self, scanning_mode):
        message = _make_test_message(
            {3: self.section_3(scanning_mode),
             6: SECTION_6_NO_BITMAP,
             7: {'codedValues': np.arange(12)}})
        data = message.data
        self.assertTrue(is_lazy_data(data))
        self.assertEqual(data.shape, (3, 4))
        self.assertEqual(data.dtype, np.floating)
        self.assertArrayEqual(as_concrete_data(data),
                              np.arange(12).reshape(3, 4))

    def test_regular_mode_0(self):
        self._test(0)

    def test_regular_mode_64(self):
        self._test(64)

    def test_regular_mode_128(self):
        self._test(128)

    def test_regular_mode_64_128(self):
        self._test(64 | 128)


def _example_section_3(grib_definition_template_number, scanning_mode):
    return {'sourceOfGridDefinition': 0,
            'numberOfOctectsForNumberOfPoints': 0,
            'interpretationOfNumberOfPoints': 0,
            'gridDefinitionTemplateNumber': grib_definition_template_number,
            'scanningMode': scanning_mode,
            'Nj': 3,
            'Ni': 4}


class Test_data__grid_template_0(tests.IrisGribTest,
                                 Mixin_data__grid_template):
    def section_3(self, scanning_mode):
        return _example_section_3(0, scanning_mode)


class Test_data__grid_template_1(tests.IrisGribTest,
                                 Mixin_data__grid_template):
    def section_3(self, scanning_mode):
        return _example_section_3(1, scanning_mode)


class Test_data__grid_template_5(tests.IrisGribTest,
                                 Mixin_data__grid_template):
    def section_3(self, scanning_mode):
        return _example_section_3(5, scanning_mode)


class Test_data__grid_template_12(tests.IrisGribTest,
                                  Mixin_data__grid_template):
    def section_3(self, scanning_mode):
        return _example_section_3(12, scanning_mode)


class Test_data__grid_template_30(tests.IrisGribTest,
                                  Mixin_data__grid_template):
    def section_3(self, scanning_mode):
        section_3 = _example_section_3(30, scanning_mode)
        # Dimensions are 'Nx' + 'Ny' instead of 'Ni' + 'Nj'.
        section_3['Nx'] = section_3['Ni']
        section_3['Ny'] = section_3['Nj']
        del section_3['Ni']
        del section_3['Nj']
        return section_3


class Test_data__grid_template_40_regular(tests.IrisGribTest,
                                          Mixin_data__grid_template):
    def section_3(self, scanning_mode):
        return _example_section_3(40, scanning_mode)


class Test_data__grid_template_90(tests.IrisGribTest,
                                  Mixin_data__grid_template):
    def section_3(self, scanning_mode):
        section_3 = _example_section_3(90, scanning_mode)
        # Exceptionally, dimensions are 'Nx' + 'Ny' instead of 'Ni' + 'Nj'.
        section_3['Nx'] = section_3['Ni']
        section_3['Ny'] = section_3['Nj']
        del section_3['Ni']
        del section_3['Nj']
        return section_3


class Test_data__grid_template_140(tests.IrisGribTest,
                                   Mixin_data__grid_template):
    def section_3(self, scanning_mode):
        section_3 = _example_section_3(140, scanning_mode)
        section_3['numberOfPointsAlongXAxis'] = section_3['Ni']
        section_3['numberOfPointsAlongYAxis'] = section_3['Nj']
        del section_3['Ni']
        del section_3['Nj']
        return section_3


class Test_data__unknown_grid_template(tests.IrisGribTest):
    def test(self):
        message = _make_test_message(
            {3: _example_section_3(999, 0),
             6: SECTION_6_NO_BITMAP,
             7: {'codedValues': np.arange(12)}})
        with self.assertRaisesRegex(TranslationError,
                                    'template 999 is not supported'):
            data = message.data


if __name__ == '__main__':
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Test function :func:`iris_grib._load_convert.product_definition_template_10`.

"""

# import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests

from copy import deepcopy
from unittest import mock

from iris.coords import DimCoord
from iris_grib._load_convert import product_definition_template_10
from iris_grib.tests.unit.load_convert import empty_metadata


class Test(tests.IrisGribTest):
    def setUp(self):
        module = 'iris_grib._load_convert'
        this_module = '{}.product_definition_template_10'.format(module)
        self.patch_statistical_fp_coord = self.patch(
            module + '.statistical_forecast_period_coord',
            return_value=mock.sentinel.dummy_fp_coord)
        self.patch_time_coord = self.patch(
            module + '.validity_time_coord',
            return_value=mock.sentinel.dummy_time_coord)
        self.patch_vertical_coords = self.patch(module + '.vertical_coords')

    def test_percentile_coord(self):
        metadata = empty_metadata()
        percentileValue = 75
        section = {'productDefinitionTemplateNumber': 10,
                   'percentileValue': percentileValue,
                   'hoursAfterDataCutoff': 1,
                   'minutesAfterDataCutoff': 1,
                   'numberOfTimeRange': 1,
                   'typeOfStatisticalProcessing': 1,
                   'typeOfTimeIncrement': 2,
                   'timeIncrement': 0,
                   'yearOfEndOfOverallTimeInterval': 2000,
                   'monthOfEndOfOverallTimeInterval': 1,
                   'dayOfEndOfOverallTimeInterval': 1,
                   'hourOfEndOfOverallTimeInterval': 1,
                   'minuteOfEndOfOverallTimeInterval': 0,
                   'secondOfEndOfOverallTimeInterval': 1}
        forecast_reference_time = mock.Mock()
        # The called being tested.
        product_definition_template_10(section, metadata,
                                       forecast_reference_time)

        expected = {'aux_coords_and_dims': []}
        percentile = DimCoord(percentileValue,
                              long_name='percentile_over_time',
                              units='no_unit')
        expected['aux_coords_and_dims'].append((percentile, None))

        self.assertEqual(metadata['aux_coords_and_dims'][-1],
                         expected['aux_coords_and_dims'][0])


if __name__ == '__main__':
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Tests for function
:func:`iris_grib._load_convert.generating_process`.

"""

# import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests

from iris_grib._load_convert import generating_process


class TestGeneratingProcess(tests.IrisGribTest):
    def setUp(self):
        self.warn_patch = self.patch('warnings.warn')

    def test_nowarn(self):
        generating_process(None)
        self.assertEqual(self.warn_patch.call_count, 0)

    def _check_warnings(self, with_forecast=True):
        module = 'iris_grib._load_convert'
        self.patch(module + '.options.warn_on_unsupported', True)
        call_args = [None]
        call_kwargs = {}
        expected_fragments = [
            'Unable to translate type of generating process',
            'Unable to translate background generating process']
        if with_forecast:
            expected_fragments.append(
                'Unable to translate forecast generating process')
        else:
            call_kwargs['include_forecast_process'] = False
        generating_process(*call_args, **call_kwargs)
        got_msgs = [call[0][0] for call in self.warn_patch.call_args_list]
        for got_msg, expected_fragment in zip(sorted(got_msgs),
                                              sorted(expected_fragments)):
            self.assertIn(expected_fragment, got_msg)

    def test_warn_full(self):
        self._check_warnings()

    def test_warn_no_forecast(self):
        self._check_warnings(with_forecast=False)


if __name__ == '__main__':
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for the :mod:`iris_grib` package."""

# import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests

import gribapi
import numpy as np
from unittest import mock

import iris

import iris_grib
from iris_grib.message import GribMessage


def _make_test_message(sections):
    raw_message = mock.Mock(sections=sections)
    recreate_raw = mock.Mock(return_value=raw_message)
    return GribMessage(raw_message, recreate_raw)


def _mock_gribapi_fetch(message, key):
    """
    Fake the gribapi key-fetch.

    Fetch key-value from the fake message (dictionary).
    If the key is not present, raise the diagnostic exception.

    """
    if key in message:
        return message[key]
    else:
        raise _mock_gribapi.GribInternalError


def _mock_gribapi__grib_is_missing(grib_message, keyname):
    """
    Fake the gribapi key-existence enquiry.

    Return whether the key exists in the fake message (dictionary).

    """
    return (keyname not in grib_message)


def _mock_gribapi__grib_get_native_type(grib_message, keyname):
    """
    Fake the gribapi type-discovery operation.

    Return type of key-value in the fake message (dictionary).
    If the key is not present, raise the diagnostic exception.

    """
    if keyname in grib_message:
        return type(grib_message[keyname])
    raise _mock_gribapi.GribInternalError(keyname)


# Construct a mock object to mimic the gribapi for GribWrapper testing.
_mock_gribapi = mock.Mock(spec=gribapi)
_mock_gribapi.GribInternalError = Exception

_mock_gribapi.grib_get_long = mock.Mock(side_effect=_mock_gribapi_fetch)
_mock_gribapi.grib_get_string = mock.Mock(side_effect=_mock_gribapi_fetch)
_mock_gribapi.grib_get_double = mock.Mock(side_effect=_mock_gribapi_fetch)
_mock_gribapi.grib_get_double_array = mock.Mock(
    side_effect=_mock_gribapi_fetch)
_mock_gribapi.grib_is_missing = mock.Mock(
    side_effect=_mock_gribapi__grib_is_missing)
_mock_gribapi.grib_get_native_type = mock.Mock(
    side_effect=_mock_gribapi__grib_get_native_type)


class FakeGribMessage(dict):
    """
    A 'fake grib message' object, for testing GribWrapper construction.

    Behaves as a dictionary, containing key-values for message keys.

    """
    def __init__(self, **kwargs):
        """
        Create a fake message object.

        General keys can be set/add as required via **kwargs.
        The 'time_code' key is specially managed.

        """
        # Start with a bare dictionary
        dict.__init__(self)
        # Extract specially-recognised keys.
        time_code = kwargs.pop('time_code', None)
        # Set the minimally required keys.
        self._init_minimal_message()
        # Also set a time-code, if given.
        if time_code is not None:
            self.set_timeunit_code(time_code)
        # Finally, add any remaining passed key-values.
        self.update(**kwargs)

    def _init_minimal_message(self):
        # Set values for all the required keys.
        self.update({
            'edition': 1,
            'Ni': 1,
            'Nj': 1,
            'numberOfValues': 1,
            'alternativeRowScanning': 0,
            'centre': 'ecmf',
            'year': 2007,
            'month': 3,
            'day': 23,
            'hour': 12,
            'minute': 0,
            'indicatorOfUnitOfTimeRange': 1,
            'shapeOfTheEarth': 6,
            'gridType': 'rotated_ll',
            'angleOfRotation': 0.0,
            'iDirectionIncrementInDegrees': 0.036,
            'jDirectionIncrementInDegrees': 0.036,
            'iScansNegatively': 0,
            'jScansPositively': 1,
            'longitudeOfFirstGridPointInDegrees': -5.70,
            'latitudeOfFirstGridPointInDegrees': -4.452,
            'jPointsAreConsecutive': 0,
            'values': np.array([[1.0]]),
            'indicatorOfParameter': 9999,
            'parameterNumber': 9999,
            'startStep': 24,
            'timeRangeIndicator': 1,
            'P1': 2, 'P2': 0,
            # time unit - needed AS WELL as 'indicatorOfUnitOfTimeRange'
            'unitOfTime': 1,
            'table2Version': 9999,
            })

    def set_timeunit_code(self, timecode):
        self['indicatorOfUnitOfTimeRange'] = timecode
        # for some odd reason, GRIB1 code uses *both* of these
        # NOTE kludge -- the 2 keys are really the same thing
        self['unitOfTime'] = timecode


class TestField(tests.IrisGribTest):
    def _test_for_coord(self, field, convert, coord_predicate, expected_points,
                        expected_bounds):
        (factories, references, standard_name, long_name, units,
         attributes, cell_methods, dim_coords_and_dims,
         aux_coords_and_dims) = convert(field)

        # Check for one and only one matching coordinate.
        coords_and_dims = dim_coords_and_dims + aux_coords_and_dims
        matching_coords = [coord for coord, _ in coords_and_dims if
                           coord_predicate(coord)]
        self.assertEqual(len(matching_coords), 1, str(matching_coords))
        coord = matching_coords[0]

        # Check points and bounds.
        if expected_points is not None:
            self.assertArrayEqual(coord.points, expected_points)

        if expected_bounds is None:
            self.assertIsNone(coord.bounds)
        else:
            self.assertArrayEqual(coord.bounds, expected_bounds)

    def assertCoordsAndDimsListsMatch(self, coords_and_dims_got,
                                      coords_and_dims_expected):
        """
        Check that coords_and_dims lists are equivalent.

        The arguments are lists of pairs of (coordinate, dimensions).
        The elements are compared one-to-one, by coordinate name (so the order
        of the lists is _not_ significant).
        It also checks that the coordinate types (DimCoord/AuxCoord) match.

        """
        def sorted_by_coordname(list):
            return sorted(list, key=lambda item: item[0].name())

        coords_and_dims_got = sorted_by_coordname(coords_and_dims_got)
        coords_and_dims_expected = sorted_by_coordname(
            coords_and_dims_expected)
        self.assertEqual(coords_and_dims_got, coords_and_dims_expected)
        # Also check coordinate type equivalences (as Coord.__eq__ does not).
        self.assertEqual(
            [type(coord) for coord, dims in coords_and_dims_got],
            [type(coord) for coord, dims in coords_and_dims_expected])


class TestGribSimple(tests.IrisGribTest):
    # A testing class that does not need the test data.
    def mock_grib(self):
        # A mock grib message, with attributes that can't be Mocks themselves.
        grib = mock.Mock()
        grib.startStep = 0
        grib.phenomenon_points = lambda unit: 3
        grib._forecastTimeUnit = "hours"
        grib.productDefinitionTemplateNumber = 0
        # define a level type (NB these 2 are effectively the same)
        grib.levelType = 1
        grib.typeOfFirstFixedSurface = 1
        grib.typeOfSecondFixedSurface = 1
        return grib

    def cube_from_message(self, grib):
        # Parameter translation now uses the GribWrapper, so we must convert
        # the Mock-based fake message to a FakeGribMessage.
        with mock.patch('iris_grib.gribapi', _mock_gribapi):
                grib_message = FakeGribMessage(**grib.__dict__)
                wrapped_msg = iris_grib.GribWrapper(grib_message)
                cube, _, _ = iris.fileformats.rules._make_cube(
                    wrapped_msg, iris_grib._grib1_load_rules.grib1_convert)
        return cube



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Test function :func:`iris_grib._load_convert.other_time_coord.

"""

# import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests

import iris.coords

from iris_grib._load_convert import other_time_coord


class TestValid(tests.IrisGribTest):
    def test_t(self):
        rt = iris.coords.DimCoord(48, 'time', units='hours since epoch')
        fp = iris.coords.DimCoord(6, 'forecast_period', units='hours')
        result = other_time_coord(rt, fp)
        expected = iris.coords.DimCoord(42, 'forecast_reference_time',
                                        units='hours since epoch')
        self.assertEqual(result, expected)

    def test_frt(self):
        rt = iris.coords.DimCoord(48, 'forecast_reference_time',
                                  units='hours since epoch')
        fp = iris.coords.DimCoord(6, 'forecast_period', units='hours')
        result = other_time_coord(rt, fp)
        expected = iris.coords.DimCoord(54, 'time', units='hours since epoch')
        self.assertEqual(result, expected)


class TestInvalid(tests.IrisGribTest):
    def test_t_with_bounds(self):
        rt = iris.coords.DimCoord(48, 'time', units='hours since epoch',
                                  bounds=[36, 60])
        fp = iris.coords.DimCoord(6, 'forecast_period', units='hours')
        with self.assertRaisesRegex(ValueError, 'bounds'):
            other_time_coord(rt, fp)

    def test_frt_with_bounds(self):
        rt = iris.coords.DimCoord(48, 'forecast_reference_time',
                                  units='hours since epoch',
                                  bounds=[42, 54])
        fp = iris.coords.DimCoord(6, 'forecast_period', units='hours')
        with self.assertRaisesRegex(ValueError, 'bounds'):
            other_time_coord(rt, fp)

    def test_fp_with_bounds(self):
        rt = iris.coords.DimCoord(48, 'time', units='hours since epoch')
        fp = iris.coords.DimCoord(6, 'forecast_period', units='hours',
                                  bounds=[3, 9])
        with self.assertRaisesRegex(ValueError, 'bounds'):
            other_time_coord(rt, fp)

    def test_vector_t(self):
        rt = iris.coords.DimCoord([0, 3], 'time', units='hours since epoch')
        fp = iris.coords.DimCoord(6, 'forecast_period', units='hours')
        with self.assertRaisesRegex(ValueError, 'Vector'):
            other_time_coord(rt, fp)

    def test_vector_frt(self):
        rt = iris.coords.DimCoord([0, 3], 'forecast_reference_time',
                                  units='hours since epoch')
        fp = iris.coords.DimCoord(6, 'forecast_period', units='hours')
        with self.assertRaisesRegex(ValueError, 'Vector'):
            other_time_coord(rt, fp)

    def test_vector_fp(self):
        rt = iris.coords.DimCoord(48, 'time', units='hours since epoch')
        fp = iris.coords.DimCoord([6, 12], 'forecast_period', units='hours')
        with self.assertRaisesRegex(ValueError, 'Vector'):
            other_time_coord(rt, fp)

    def test_invalid_rt_name(self):
        rt = iris.coords.DimCoord(1, 'height')
        fp = iris.coords.DimCoord(6, 'forecast_period', units='hours')
        with self.assertRaisesRegex(ValueError, 'reference time'):
            other_time_coord(rt, fp)

    def test_invalid_t_unit(self):
        rt = iris.coords.DimCoord(1, 'time', units='Pa')
        fp = iris.coords.DimCoord(6, 'forecast_period', units='hours')
        with self.assertRaisesRegex(ValueError, 'unit.*Pa'):
            other_time_coord(rt, fp)

    def test_invalid_frt_unit(self):
        rt = iris.coords.DimCoord(1, 'forecast_reference_time', units='km')
        fp = iris.coords.DimCoord(6, 'forecast_period', units='hours')
        with self.assertRaisesRegex(ValueError, 'unit.*km'):
            other_time_coord(rt, fp)

    def test_invalid_fp_unit(self):
        rt = iris.coords.DimCoord(48, 'forecast_reference_time',
                                  units='hours since epoch')
        fp = iris.coords.DimCoord(6, 'forecast_period', units='kg')
        with self.assertRaisesRegex(ValueError, 'unit.*kg'):
            other_time_coord(rt, fp)


if __name__ == '__main__':
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Test function
:func:`iris_grib._load_convert.product_definition_template_1`.

"""

# import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests

from copy import deepcopy
from unittest import mock
import warnings

from iris.coords import DimCoord

from iris_grib._load_convert import product_definition_template_1


class Test(tests.IrisGribTest):
    def setUp(self):
        def func(s, m, f):
            return m['cell_methods'].append(self.cell_method)

        module = 'iris_grib._load_convert'
        self.patch('warnings.warn')
        this = '{}.product_definition_template_0'.format(module)
        self.cell_method = mock.sentinel.cell_method
        self.patch(this, side_effect=func)
        self.metadata = {'factories': [], 'references': [],
                         'standard_name': None,
                         'long_name': None, 'units': None, 'attributes': {},
                         'cell_methods': [], 'dim_coords_and_dims': [],
                         'aux_coords_and_dims': []}

    def _check(self, request_warning):
        this = 'iris_grib._load_convert.options'
        with mock.patch(this, warn_on_unsupported=request_warning):
            metadata = deepcopy(self.metadata)
            perturbationNumber = 666
            section = {'perturbationNumber': perturbationNumber}
            forecast_reference_time = mock.sentinel.forecast_reference_time
            # The called being tested.
            product_definition_template_1(section, metadata,
                                          forecast_reference_time)
            expected = deepcopy(self.metadata)
            expected['cell_methods'].append(self.cell_method)
            realization = DimCoord(perturbationNumber,
                                   standard_name='realization',
                                   units='no_unit')
            expected['aux_coords_and_dims'].append((realization, None))
            self.assertEqual(metadata, expected)
            if request_warning:
                warn_msgs = [mcall[1][0] for mcall in warnings.warn.mock_calls]
                expected_msgs = ['type of ensemble', 'number of forecasts']
                for emsg in expected_msgs:
                    matches = [wmsg for wmsg in warn_msgs if emsg in wmsg]
                    self.assertEqual(len(matches), 1)
                    warn_msgs.remove(matches[0])
            else:
                self.assertEqual(len(warnings.warn.mock_calls), 0)

    def test_pdt_no_warn(self):
        self._check(False)

    def test_pdt_warn(self):
        self._check(True)


if __name__ == '__main__':
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for the `iris_grib.message._RawGribMessage` class.

"""

# Import iris_grib.tests first so that some things can be initialised before
# importing anything else.
import iris_grib.tests as tests

import gribapi

from iris_grib.message import _RawGribMessage


@tests.skip_data
class Test(tests.IrisGribTest):
    def setUp(self):
        filename = tests.get_data_path(('GRIB', 'uk_t', 'uk_t.grib2'))
        with open(filename, 'rb') as grib_fh:
            grib_id = gribapi.grib_new_from_file(grib_fh)
            self.message = _RawGribMessage(grib_id)

    def test_sections__set(self):
        # Test that sections writes into the _sections attribute.
        res = self.message.sections
        self.assertNotEqual(self.message._sections, None)

    def test_sections__indexing(self):
        res = self.message.sections[3]['scanningMode']
        expected = 64
        self.assertEqual(expected, res)

    def test__get_message_sections__section_numbers(self):
        res = list(self.message.sections.keys())
        self.assertEqual(res, list(range(9)))

    def test_sections__numberOfSection_value(self):
        # The key `numberOfSection` is repeated in every section meaning that
        # if requested using gribapi it always defaults to its last value (7).
        # This tests that the `_RawGribMessage._get_message_sections`
        # override is functioning.
        section_number = 4
        res = self.message.sections[section_number]['numberOfSection']
        self.assertEqual(res, section_number)


if __name__ == '__main__':
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for iris-grib pep8 conformance.

"""

import os
import unittest

import pep8

import iris_grib


class Test(unittest.TestCase):
    def test_pep8_conformance(self):
        pep8style = pep8.StyleGuide(quiet=False)
        excluded = ['_grib_cf_map.py', '_grib1_load_rules.py']
        for fname in excluded:
            path = '*{}{}'.format(os.path.sep, fname)
            pep8style.options.exclude.append(path)

        extra_exclude_fname = os.path.join(os.path.dirname(__file__),
                                           '.pep8_test_exclude.txt')

        if os.path.exists(extra_exclude_fname):
            with open(extra_exclude_fname) as fh:
                extra_exclude = [line.strip() for line in fh if line.strip()]
            pep8style.options.exclude.extend(extra_exclude)

        root = os.path.dirname(os.path.abspath(iris_grib.__file__))
        result = pep8style.check_files([root])
        emsg = 'Found code pep8 errors (and warnings).'
        self.assertEqual(result.total_errors, 0, emsg)


if __name__ == '__main__':
    unittest.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Test function
:func:`iris_grib._load_convert.grid_definition_template_0_and_1`.

"""

# Import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests

import numpy as np

import iris.coord_systems
import iris.coords

from iris.exceptions import TranslationError

from iris_grib.tests.unit.load_convert import empty_metadata

from iris_grib._load_convert import grid_definition_template_0_and_1


class _Section(dict):
    def get_computed_key(self, key):
        return self.get(key)


class Test_resolution_flags(tests.IrisGribTest):

    def section_3(self):
        section = _Section({
            'Ni': 6,
            'Nj': 6,
            'latitudeOfFirstGridPoint': 0,
            'longitudeOfFirstGridPoint': 0,
            'resolutionAndComponentFlags': 0,
            'latitudeOfLastGridPoint': 5000000,
            'longitudeOfLastGridPoint': 5000000,
            'iDirectionIncrement': 0,
            'jDirectionIncrement': 0,
            'scanningMode': 0b01000000,
            'numberOfOctectsForNumberOfPoints': 0,
            'interpretationOfNumberOfPoints': 0,
        })
        return section

    def expected(self, x_dim, y_dim, x_points, y_points, x_neg=True,
                 y_neg=True):
        # Prepare the expectation.
        expected = empty_metadata()
        cs = iris.coord_systems.GeogCS(6367470)
        if x_neg:
            x_points = x_points[::-1]
        x = iris.coords.DimCoord(x_points,
                                 standard_name='longitude',
                                 units='degrees',
                                 coord_system=cs)
        if y_neg:
            y_points = y_points[::-1]
        y = iris.coords.DimCoord(y_points,
                                 standard_name='latitude',
                                 units='degrees',
                                 coord_system=cs)
        expected['dim_coords_and_dims'].append((y, y_dim))
        expected['dim_coords_and_dims'].append((x, x_dim))
        return expected

    def test_without_increments(self):
        section = self.section_3()
        metadata = empty_metadata()
        cs = iris.coord_systems.GeogCS(6367470)
        grid_definition_template_0_and_1(section, metadata, 'latitude',
                                         'longitude', cs)
        x_points = np.array([0., 1., 2., 3., 4., 5.])
        y_points = np.array([0., 1., 2., 3., 4., 5.])
        expected = self.expected(1, 0, x_points, y_points, x_neg=False,
                                 y_neg=False)
        self.assertEqual(metadata, expected)

    def test_with_increments(self):
        section = self.section_3()
        section['resolutionAndComponentFlags'] = 48
        section['iDirectionIncrement'] = 1000000
        section['jDirectionIncrement'] = 1000000
        metadata = empty_metadata()
        cs = iris.coord_systems.GeogCS(6367470)
        grid_definition_template_0_and_1(section, metadata, 'latitude',
                                         'longitude', cs)
        x_points = np.array([0., 1., 2., 3., 4., 5.])
        y_points = np.array([0., 1., 2., 3., 4., 5.])
        expected = self.expected(1, 0, x_points, y_points, x_neg=False,
                                 y_neg=False)
        self.assertEqual(metadata, expected)

    def test_with_i_not_j_increment(self):
        section = self.section_3()
        section['resolutionAndComponentFlags'] = 32
        section['iDirectionIncrement'] = 1000000
        metadata = empty_metadata()
        cs = iris.coord_systems.GeogCS(6367470)
        grid_definition_template_0_and_1(section, metadata, 'latitude',
                                         'longitude', cs)
        x_points = np.array([0., 1., 2., 3., 4., 5.])
        y_points = np.array([0., 1., 2., 3., 4., 5.])
        expected = self.expected(1, 0, x_points, y_points, x_neg=False,
                                 y_neg=False)
        self.assertEqual(metadata, expected)

    def test_with_j_not_i_increment(self):
        section = self.section_3()
        section['resolutionAndComponentFlags'] = 16
        section['jDirectionIncrement'] = 1000000
        metadata = empty_metadata()
        cs = iris.coord_systems.GeogCS(6367470)
        grid_definition_template_0_and_1(section, metadata, 'latitude',
                                         'longitude', cs)
        x_points = np.array([0., 1., 2., 3., 4., 5.])
        y_points = np.array([0., 1., 2., 3., 4., 5.])
        expected = self.expected(1, 0, x_points, y_points, x_neg=False,
                                 y_neg=False)
        self.assertEqual(metadata, expected)

    def test_without_increments_crossing_0_lon(self):
        section = self.section_3()
        section['longitudeOfFirstGridPoint'] = 355000000
        section['Ni'] = 11
        metadata = empty_metadata()
        cs = iris.coord_systems.GeogCS(6367470)
        grid_definition_template_0_and_1(section, metadata, 'latitude',
                                         'longitude', cs)
        x_points = np.array(
            [355., 356., 357., 358., 359., 360., 361., 362., 363., 364., 365.]
        )
        y_points = np.array([0., 1., 2., 3., 4., 5.])
        expected = self.expected(1, 0, x_points, y_points, x_neg=False,
                                 y_neg=False)
        self.assertEqual(metadata, expected)


class Test(tests.IrisGribTest):

    def test_unsupported_quasi_regular__number_of_octets(self):
        section = {'numberOfOctectsForNumberOfPoints': 1}
        cs = None
        metadata = None
        with self.assertRaisesRegex(TranslationError, 'quasi-regular'):
            grid_definition_template_0_and_1(section,
                                             metadata,
                                             'latitude',
                                             'longitude',
                                             cs)

    def test_unsupported_quasi_regular__interpretation(self):
        section = {'numberOfOctectsForNumberOfPoints': 1,
                   'interpretationOfNumberOfPoints': 1}
        cs = None
        metadata = None
        with self.assertRaisesRegex(TranslationError, 'quasi-regular'):
            grid_definition_template_0_and_1(section,
                                             metadata,
                                             'latitude',
                                             'longitude',
                                             cs)


if __name__ == '__main__':
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for
:func:`iris_grib._save_rules._product_definition_template_8_10_and_11`

"""

# Import iris_grib.tests first so that some things can be initialised before
# importing anything else.
import iris_grib.tests as tests

from unittest import mock

from cf_units import Unit
import gribapi

from iris.coords import CellMethod, DimCoord
import iris.tests.stock as stock

from iris_grib._save_rules import _product_definition_template_8_10_and_11


class TestTypeOfStatisticalProcessing(tests.IrisGribTest):
    def setUp(self):
        self.cube = stock.lat_lon_cube()
        # Rename cube to avoid warning about unknown discipline/parameter.
        self.cube.rename('air_temperature')
        coord = DimCoord(23, 'time', bounds=[0, 100],
                         units=Unit('days since epoch', calendar='standard'))
        self.cube.add_aux_coord(coord)

    @mock.patch.object(gribapi, 'grib_set')
    def test_sum(self, mock_set):
        cube = self.cube
        cell_method = CellMethod(method='sum', coords=['time'])
        cube.add_cell_method(cell_method)

        _product_definition_template_8_10_and_11(cube, mock.sentinel.grib)
        mock_set.assert_any_call(mock.sentinel.grib,
                                 "typeOfStatisticalProcessing", 1)

    @mock.patch.object(gribapi, 'grib_set')
    def test_unrecognised(self, mock_set):
        cube = self.cube
        cell_method = CellMethod(method='95th percentile', coords=['time'])
        cube.add_cell_method(cell_method)

        _product_definition_template_8_10_and_11(cube, mock.sentinel.grib)
        mock_set.assert_any_call(mock.sentinel.grib,
                                 "typeOfStatisticalProcessing", 255)

    @mock.patch.object(gribapi, 'grib_set')
    def test_multiple_cell_method_coords(self, mock_set):
        cube = self.cube
        cell_method = CellMethod(method='sum',
                                 coords=['time', 'forecast_period'])
        cube.add_cell_method(cell_method)
        with self.assertRaisesRegex(ValueError,
                                    'Cannot handle multiple coordinate name'):
            _product_definition_template_8_10_and_11(cube, mock.sentinel.grib)

    @mock.patch.object(gribapi, 'grib_set')
    def test_cell_method_coord_name_fail(self, mock_set):
        cube = self.cube
        cell_method = CellMethod(method='mean', coords=['season'])
        cube.add_cell_method(cell_method)
        with self.assertRaisesRegex(
                ValueError, "Expected a cell method with a coordinate "
                "name of 'time'"):
            _product_definition_template_8_10_and_11(cube, mock.sentinel.grib)


class TestTimeCoordPrerequisites(tests.IrisGribTest):
    def setUp(self):
        self.cube = stock.lat_lon_cube()
        # Rename cube to avoid warning about unknown discipline/parameter.
        self.cube.rename('air_temperature')

    @mock.patch.object(gribapi, 'grib_set')
    def test_multiple_points(self, mock_set):
        # Add time coord with multiple points.
        coord = DimCoord([23, 24, 25], 'time',
                         bounds=[[22, 23], [23, 24], [24, 25]],
                         units=Unit('days since epoch', calendar='standard'))
        self.cube.add_aux_coord(coord, 0)
        with self.assertRaisesRegex(
                ValueError, 'Expected length one time coordinate'):
            _product_definition_template_8_10_and_11(self.cube,
                                                     mock.sentinel.grib)

    @mock.patch.object(gribapi, 'grib_set')
    def test_no_bounds(self, mock_set):
        # Add time coord with no bounds.
        coord = DimCoord(23, 'time',
                         units=Unit('days since epoch', calendar='standard'))
        self.cube.add_aux_coord(coord)
        with self.assertRaisesRegex(
                ValueError, 'Expected time coordinate with two bounds, '
                'got 0 bounds'):
            _product_definition_template_8_10_and_11(self.cube,
                                                     mock.sentinel.grib)

    @mock.patch.object(gribapi, 'grib_set')
    def test_more_than_two_bounds(self, mock_set):
        # Add time coord with more than two bounds.
        coord = DimCoord(23, 'time', bounds=[21, 22, 23],
                         units=Unit('days since epoch', calendar='standard'))
        self.cube.add_aux_coord(coord)
        with self.assertRaisesRegex(
                ValueError, 'Expected time coordinate with two bounds, '
                'got 3 bounds'):
            _product_definition_template_8_10_and_11(self.cube,
                                                     mock.sentinel.grib)


class TestEndOfOverallTimeInterval(tests.IrisGribTest):
    def setUp(self):
        self.cube = stock.lat_lon_cube()
        # Rename cube to avoid warning about unknown discipline/parameter.
        self.cube.rename('air_temperature')
        cell_method = CellMethod(method='sum', coords=['time'])
        self.cube.add_cell_method(cell_method)

    @mock.patch.object(gribapi, 'grib_set')
    def test_default_calendar(self, mock_set):
        cube = self.cube
        # End bound is 1972-04-26 10:27:07.
        coord = DimCoord(23.0, 'time', bounds=[0.452, 20314.452],
                         units=Unit('hours since epoch'))
        cube.add_aux_coord(coord)

        grib = mock.sentinel.grib
        _product_definition_template_8_10_and_11(cube, grib)

        mock_set.assert_any_call(
            grib, "yearOfEndOfOverallTimeInterval", 1972)
        mock_set.assert_any_call(
            grib, "monthOfEndOfOverallTimeInterval", 4)
        mock_set.assert_any_call(
            grib, "dayOfEndOfOverallTimeInterval", 26)
        mock_set.assert_any_call(
            grib, "hourOfEndOfOverallTimeInterval", 10)
        mock_set.assert_any_call(
            grib, "minuteOfEndOfOverallTimeInterval", 27)
        mock_set.assert_any_call(
            grib, "secondOfEndOfOverallTimeInterval", 7)

    @mock.patch.object(gribapi, 'grib_set')
    def test_360_day_calendar(self, mock_set):
        cube = self.cube
        # End bound is 1972-05-07 10:27:07
        coord = DimCoord(23.0, 'time', bounds=[0.452, 20314.452],
                         units=Unit('hours since epoch', calendar='360_day'))
        cube.add_aux_coord(coord)

        grib = mock.sentinel.grib
        _product_definition_template_8_10_and_11(cube, grib)

        mock_set.assert_any_call(
            grib, "yearOfEndOfOverallTimeInterval", 1972)
        mock_set.assert_any_call(
            grib, "monthOfEndOfOverallTimeInterval", 5)
        mock_set.assert_any_call(
            grib, "dayOfEndOfOverallTimeInterval", 7)
        mock_set.assert_any_call(
            grib, "hourOfEndOfOverallTimeInterval", 10)
        mock_set.assert_any_call(
            grib, "minuteOfEndOfOverallTimeInterval", 27)
        mock_set.assert_any_call(
            grib, "secondOfEndOfOverallTimeInterval", 7)


class TestNumberOfTimeRange(tests.IrisGribTest):
    @mock.patch.object(gribapi, 'grib_set')
    def test_other_cell_methods(self, mock_set):
        cube = stock.lat_lon_cube()
        # Rename cube to avoid warning about unknown discipline/parameter.
        cube.rename('air_temperature')
        coord = DimCoord(23, 'time', bounds=[0, 24],
                         units=Unit('hours since epoch'))
        cube.add_aux_coord(coord)
        # Add one time cell method and another unrelated one.
        cell_method = CellMethod(method='mean', coords=['elephants'])
        cube.add_cell_method(cell_method)
        cell_method = CellMethod(method='sum', coords=['time'])
        cube.add_cell_method(cell_method)

        _product_definition_template_8_10_and_11(cube, mock.sentinel.grib)
        mock_set.assert_any_call(mock.sentinel.grib, 'numberOfTimeRange', 1)


if __name__ == "__main__":
    tests.main()
